/*
ChrispyChars AI
	by Lach

v1.2
*/

local CC = ChrispyChars
local DELAYS = CC.BOTDELAYS

local function valid(mo)
	return mo and mo.valid
end

function CC.CreateBotTable()
	local leveltime = leveltime
	return {
		flags = leveltime,
		dabTic = leveltime,
		crouchTic = leveltime,
		emoteTic = leveltime,
		wheelTic = leveltime
	}
end

function CC.UpdateBot(player)
	local chrispy = player.chrispy
	local leader = CC.GetBotLeader(player)
	
	if leader == nil
		return
	end
	
	// restat bot if they acquire a new bot leader
	if player.botleader ~= chrispy.botleader
		chrispy.botleader = player.botleader
		CC.SetStats(player)
	end
	
	// copy ChrispyChars toggles
	for _, var in ipairs(CC.BOT_COPYVARS)
		player[var] = leader[var]
	end
	
	// copy powers
	for _, pow in ipairs(CC.BOT_COPYPOWERS)
		if leader.powers[pow]
		and not player.powers[pow]
			player.powers[pow] = leader.powers[pow]
		end
	end
	
	// copy super status
	if leader.powers[pw_super]
		player.rings = leader.rings
		if not player.powers[pw_super]
			local mo = leader.mo
			if mo and mo.state >= S_PLAY_SUPER_TRANS1
			and mo.state <= S_PLAY_SUPER_TRANS6
				P_DoSuperTransformation(player)
			else
				player.powers[pw_super] = $ + 1
			end
			chrispy.flags = $ | CF_SUPERBOT
		end
	elseif chrispy.flags & CF_SUPERBOT
		chrispy.flags = $ & ~CF_SUPERBOT
		player.rings = 0
	end
end

CC.AddHook("BotTiccmd", function(player2, cmd2)
	local player1 = CC.GetBotLeader(player2)
	
	if not player1
		return
	end
	
	local cmd1 = player1.cmd
	local mo1 = player1.mo
	local mo2 = player2.mo
	local chrispy1 = player1.chrispy
	local chrispy2 = player2.chrispy
	
	if player1.exiting == 1
	or not (valid(mo1) and valid(mo2))
	or not (chrispy1 and chrispy2)
		return
	end
	
	if player2.powers[pw_carry] == CR_DUSTDEVIL
	or player2.powers[pw_carry] == CR_PLAYER
		return true
	end
	
	local leveltime = leveltime
	local bot = chrispy2.bot
	
	if not bot
		bot = CC.CreateBotTable()
		chrispy2.bot = bot
	end
	
	if P_IsObjectOnGround(mo2)
		bot.flags = $ & ~CF_FLIPPED
	end
	
	// don't cancel a high five!
	if mo2.state == S_CHRISPY_HIGHFIVE
	and CC.ActiveHighFiveState(player2)
		return true
	end
	
	// conditions for actions
	
	local highFive = mo1.state == S_CHRISPY_HIGHFIVE and chrispy1.highFiveState == 0
	local tryCrouch
	
	if highFive
		tryCrouch = not (chrispy2.flags & CF_CROUCHING)
	else
		tryCrouch = (chrispy1.flags & CF_CROUCHING) ~= (chrispy2.flags & CF_CROUCHING)
	end
	
	// looking up/down
	
	if twodlevel or mo2.flags2 & MF2_TWOD
	or P_AproxDistance(mo1.x - mo2.x, mo1.y - mo2.y) < 32*mo2.scale
		cmd2.aiming = cmd1.aiming
	end
	
	// Fly with momentum flip
	
	if chrispy1.flags & CF_FLIPPED
	and player2.charability == CA_FLY
	and player2.pflags & (PF_JUMPED|PF_THOKKED) == PF_JUMPED
	and not (mo2.eflags & MFE_UNDERWATER)
	and chrispy2.time & 1
		cmd2.buttons = $ | BT_JUMP // mash jump!
		bot.flags = $ | CF_FLIPPED
		if TCF_SETCARRY // in case !1679 gets merged
			cmd2.flags = $ | TCF_SETCARRY
		end
	end
	
	// Drop dash
	
	if chrispy1.flags & CF_DROPDASHCHARGE
	and player2.pflags & PF_JUMPED
		bot.flags = $ | CF_DROPDASHCHARGE
	end
	
	if bot.flags & CF_DROPDASHCHARGE
		if player2.pflags & PF_JUMPED
			cmd2.buttons = $ | BT_USE
		end
		if not (chrispy1.flags & CF_DROPDASHCHARGE or player1.pflags & PF_SPINNING)
			bot.flags = $ & ~CF_DROPDASHCHARGE
		end
	end
	
	// Crouch
	
	if tryCrouch
	and (player2.panim == PA_IDLE or mo2.state == S_PLAY_LOOK)
		if leveltime - bot.crouchTic > DELAYS.CROUCH
		and chrispy2.time & 1
			cmd2.buttons = $ | CC_DEFAULTBUTTONS.crouch
		end
	else
		bot.crouchTic = leveltime
	end
	
	// Emote wheel
	
	if chrispy2.flags & CF_CROUCHING
		if chrispy2.flags & CF_EMOTEWHEEL
			if not highFive
				if chrispy2.time & 1
					cmd2.buttons = $ | CC_DEFAULTBUTTONS.dance
				end
			else
				local emotewheel = chrispy2.emotewheel
				local emoteNum = CC.emotes.dictionary.highfive
				if emotewheel.selection == emoteNum
					if leveltime - bot.emoteTic > DELAYS.EMOTE
					and chrispy2.time & 1
						cmd2.buttons = $ | BT_JUMP
					end
				elseif emotewheel.left >= 0
					bot.emoteTic = leveltime
					if emotewheel.selection - emoteNum > 0
						cmd2.sidemove = -50
					else
						cmd2.sidemove = 50
					end
				end
			end
			return true
		end
		
		if highFive
			if leveltime - bot.wheelTic > DELAYS.EMOTE
			and chrispy2.time & 1
				cmd2.buttons = $ | CC_DEFAULTBUTTONS.dance
				bot.emoteTic = leveltime
			end
		else
			bot.wheelTic = leveltime
		end
	end
	
	// Dab (yeah sorry this block is a bit barbaric)
	
	if mo2.state == S_CHRISPY_DAB // if dabbing, undab after a short period of time
		if mo1.state == S_CHRISPY_DAB
			bot.dabTic = leveltime
			cmd2.forwardmove = 0 // don't move!
			cmd2.sidemove = 0
			return true
		else
			if leveltime - bot.dabTic > DELAYS.DAB_EXIT
			and chrispy2.time & 1
				cmd2.buttons = $ | CC_DEFAULTBUTTONS.dab
			end
		end
	elseif mo1.state == S_CHRISPY_DAB // if not dabbing, but player 1 is dabbing, try to dab
		if mo2.state == S_PLAY_LOOK // if neck is enabled, wait a little while to react to player 1 first
		and player2.emotion_enabled ~= false
			if mo2.neck and mo2.neck.target == mo1
				if leveltime - bot.dabTic > DELAYS.DAB_ENTER
				and mo2.state ~= S_CHRISPY_DAB
					cmd2.buttons = $ | CC_DEFAULTBUTTONS.dab
				end
			end
		elseif mo2.state ~= S_CHRISPY_DAB
		and chrispy2.time & 1 // MASH THE DAB BUTTON
			cmd2.buttons = $ | CC_DEFAULTBUTTONS.dab
		end
	else
		bot.dabTic = leveltime
	end
end)

if not TCF_SETCARRY // in case !1679 gets merged
	CC.AddHook("AbilitySpecial", function(player)
		local chrispy = player.chrispy
		local bot = chrispy and chrispy.bot
		
		if player.pflags & PF_THOKKED
		or not (bot and bot.flags & CF_FLIPPED)
			return
		end
		
		bot.flags = $ & ~CF_FLIPPED
		player.pflags = $ | PF_CANCARRY
	end)
end

CC.AddHook("PlayerThink", function(player)
	local chrispy = player.chrispy
	
	if not chrispy then return end
	
	if player.playerstate == PST_REBORN
	and player.bot ~= CC.BOT_NONE
	and player.powers[pw_shield] ~= 0
		chrispy.transfershield = player.powers[pw_shield]
	end
end)
