/*
ChrispyChars emotes
	by Lach

v1.2
*/

local CC = ChrispyChars

CC.EMOTEFLAGS = {}
CC.EMOTEFLAGS.SETSTATE = 1 << 0
CC.EMOTEFLAGS.REQUIREGROUND = 1 << 1
CC.EMOTEFLAGS.REQUIRESTATIONARY = 1 << 2
CC.EMOTEFLAGS.REQUIREALL = CC.EMOTEFLAGS.REQUIREGROUND | CC.EMOTEFLAGS.REQUIRESTATIONARY

CC.emotes = {
	{
		tags = {"highfive", "hello", "wave"},
		flags = CC.EMOTEFLAGS.SETSTATE|CC.EMOTEFLAGS.REQUIREALL,
		state = S_CHRISPY_HIGHFIVE,
		func = function(player)
			player.chrispy.highFiveState = 0
		end,
	},
	{
		tags = {"squiggle", "annoyed", "sad", "respects"},
		flags = CC.EMOTEFLAGS.REQUIREALL,
		state = S_CHRISPY_HATER,
		tick = function(player)
			return skins[player.skin].contspeed
		end
	},
	{
		tags = {"question", "curious"},
		flags = CC.EMOTEFLAGS.REQUIREALL,
		state = S_PLAY_EDGE
	},
	{
		tags = {"exclamation", "danger", "alert"},
		state = S_PLAY_SKID
	},
	{
		tags = {"heart", "love"},
		flags = CC.EMOTEFLAGS.REQUIREALL,
		state = S_PLAY_WALK
	},
	{
		tags = {"chaojoy", "laugh", "rofl"},
		flags = CC.EMOTEFLAGS.REQUIREALL,
		state = S_PLAY_ROLL
	},
	{
		tags = {"tpose"},
		flags = CC.EMOTEFLAGS.SETSTATE,
		state = S_CHRISPY_TPOSE,
	},
	{
		tags = {"surprised", "shocked", "pog", "sonic"},
		state = S_PLAY_GASP
	},
	{
		tags = {"thinking", "menu", "tails"},
		flags = CC.EMOTEFLAGS.REQUIREALL,
		state = S_PLAY_WAIT
	},
	{
		tags = {"fangmoney", "cash"},
		flags = CC.EMOTEFLAGS.REQUIREALL,
		state = S_PLAY_RIDE,
		func = function(player)
			CC_SpawnConfetti(player, 1, 12*FRACUNIT, nil, false, function(part)
				A_SetObjectFlags(part, MF_NOBLOCKMAP, 1)
				part.state = S_CHRISPY_CASH
				part.flags2 = $ | MF2_SKULLFLY
			end)
		end,
		statecondition = function(player)
			return player.cc_neckenabled == false
			or player.cc_curiosityenabled == false
			or player.cc_emotionenabled == false
		end
	},
	{
		tags = {"hater", "shadow", "angry"},
		flags = CC.EMOTEFLAGS.SETSTATE|CC.EMOTEFLAGS.REQUIREALL,
		state = S_CHRISPY_HATER
	},
	{
		tags = {"eggman", "dead"},
		sprite2 = SPR2_FROZ
	},
	{
		tags = {"okhand", "nice"},
		state = S_PLAY_STND
	},
	{
		tags = {"clap", "applause"},
		state = S_PLAY_STND
	}
}

CC.emotes.dictionary = {}

// Emote wheel table setup
for i, item in ipairs(CC.emotes)
	local found = false
	for j, trackable in ipairs(CC_NeckSkins["pointy"].trackables)
		local tag = item.tags[1]
		if trackable[tag]
			CC.emotes[tag] = j
			CC.emotes.dictionary[tag] = i
			item.id = j
			found = true
			break
		end
	end
	assert(found, "No wheel emote found for identifier "..item.id)
end

CC.numEmotes = #CC.emotes

CC.GetEmoteWheelAngle = function(selection)
	return FixedAngle((selection - 1)*360*FRACUNIT/CC.numEmotes)
end

CC.StartEmote = function(player)
	local mo = player.mo
	local chrispy = player.chrispy
	local emotewheel = chrispy.emotewheel
	local data = CC.emotes[emotewheel.selection]
	local flags = data.flags or 0
	
	if flags & CC.EMOTEFLAGS.SETSTATE
		mo.state = data.state
	elseif (data.state
	or data.sprite2 ~= nil)
	and (not data.statecondition or data.statecondition(player))
		local sprite2
		if data.state
			sprite2 = states[data.state].frame
		else
			sprite2 = data.sprite2
		end
		
		while not P_IsValidSprite2(mo, sprite2)
			sprite2 = spr2defaults[sprite2]
		end
		if player.powers[pw_super] and not (player.charflags & SF_NOSUPERSPRITES)
		and P_IsValidSprite2(mo, sprite2 | FF_SPR2SUPER)
			sprite2 = $ | FF_SPR2SUPER
		end
		
		mo.state = S_CHRISPY_ACT
		mo.sprite2 = sprite2
		
		emotewheel.state = data.state
		if data.state
			local state = states[data.state]
			if state.frame & FF_ANIMATE
				emotewheel.duration = state.var2
			else
				emotewheel.duration = state.tics
			end
		else
			emotewheel.duration = -1
		end
		emotewheel.tics = data.tick and data.tick(player) or emotewheel.duration
	else
		mo.state = S_PLAY_STND
	end
	
	if data.func
		data.func(player)
	end
	
	chrispy.flags = $ & ~CF_EMOTEWHEEL
end

// emote command, for hotkeying emotes

CC.Commands = {} // this probably shouldn't be initialized here, but I really do not care about moving the stuff from MENU.lua to it just yet :V

CC.Commands.Emote = function(player, arg)
	local printPlayer = splitscreen and consoleplayer or player
	
	if not arg
		local desc = "Shortcut command for emoting via ChrispyChars' emote wheel."
			+ "\nUse with keybinds for maximum efficiency!"
			+ "\nAvailable emotes are:"
		for i, data in ipairs(CC.emotes)
			desc = $ .. "\n" .. (i < 10 and " " .. i or i) .. ": "
			for j, tag in ipairs(data.tags)
				desc = $ .. tag .. ", "
			end
			desc = $:sub(1, $:len() - 2)
		end
		CONS_Printf(printPlayer, desc)
		return
	end
	
	// make sure we can emote at all
	if not CC.CanEmote(player)
		return
	end
	
	local mo = player.mo
	local chrispy = player.chrispy
	local emotewheel = chrispy.emotewheel
	
	local num = tonumber(arg)
	
	if num ~= nil // emote number was entered directly
		if not CC.emotes[num]
			return
		end
	else // find the best-matching string
		local matches = {}
		local numMatches
		
		arg = $:gsub(" ", ""):lower()
		
		for i, data in ipairs(CC.emotes)
			for j, tag in ipairs(data.tags)
				local len = arg:len()
				local arg = arg
				
				while len
					local startPos, endPos = tag:find(arg)
					
					if startPos
						table.insert(matches, {index = i, start = startPos, length = endPos - startPos})
						break
					end
					len = $ - 1
					arg = $:sub(1, len)
				end
			end
		end
		
		numMatches = #matches
		
		if not numMatches // no emotes matched your term
			return
		end
		
		if numMatches == 1 // don't bother sorting if there's only one possible emote
			num = matches[1].index
		else
			table.sort(matches, function(a, b)
				if a.length == b.length
					return a.start < b.start
				end
				return a.length > b.length
			end)
			num = matches[1].index
		end
	end
	
	// check that we meet further requirements for this emote
	
	local flags = CC.emotes[num].flags or 0
	
	if flags & CC.EMOTEFLAGS.REQUIREGROUND
	and not P_IsObjectOnGround(mo)
		return
	end
	
	if flags & CC.EMOTEFLAGS.REQUIRESTATIONARY
	and player.speed ~= 0
		return
	end
	
	// use the emote!
	
	if mo.state ~= S_CHRISPY_ACT
	or emotewheel.selection ~= num
		emotewheel.selection = num
		emotewheel.angle = CC.GetEmoteWheelAngle(num)
		CC.StartEmote(player)
	end
end

COM_AddCommand("emote2", CC.Commands.Emote, COM_SPLITSCREEN)
COM_AddCommand("emote", CC.Commands.Emote)