/*
ChrispyChars functions
	by Lach

v1.2
*/

local CC = ChrispyChars

function CC.Print(...)
	local str = ...
	if str == nil
		return
	end
	
	print(CC.DEBUGPREFIX .. str)
	CC.Print(select(2, ...))
end

function CC.Valid(mo)
	return mo and mo.valid
end

// identical to S_StartSound, but any sound played for a bot player is played for the lead player instead
function CC.StartSound(mo, sound, player)
	if player ~= nil
		player = CC.GetBotLeader(player) or $
	end
	
	S_StartSound(mo, sound, player)
end

// returns the shield properties for @shieldType
CC.GetShieldProperties = function(shieldType)
	return CC.shieldProperties[shieldType] or 0
end

// spawns spindash-esque dust
CC.SpawnDust = function(player, dash, destscale, offset, num, x, y, z)
	local mo = player.mo
	local chrispy = player.chrispy
	local thrust = -8*dash
	
	num = $ or 1
	
	if x == nil
		x = mo.x
	end
	if y == nil
		y = mo.y
	end
	if z == nil
		z = mo.z
	end
	
	for i = 1, num
		local angle = player.drawangle + P_RandomRange(-20, 20) * ANG1
		local xthrust = FixedMul(thrust + offset, cos(angle))
		local ythrust = FixedMul(thrust + offset, sin(angle))
		local zthrust = FixedMul(4 * dash, (FRACUNIT >> 1) + abs(sin(chrispy.time * ANG15)))
		
		local dust = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_CHRISPY_DUST)
		P_TeleportMove(dust,
			x + xthrust,
			y + ythrust,
			z + dust.z - mo.z)
		dust.target = mo
		dust.destscale = destscale
		dust.scalespeed = FixedMul($ >> 1, mo.scale)
		dust.angle = angle
		P_InstaThrust(dust, angle, thrust)
		dust.momz = P_MobjFlip(mo)*zthrust
	end
end

// spawns an angled dust ring based on @mo's velocity
// @speed sets how fast the particles move away from @mo in the direction of the dust plane
// @thrust sets how fast the particles move away from @mo in the direction perpedicular to the dust plane
// if @alwaysabove is truthy, the particles will always pretend @mo's momz is negative, so they always spawn "above" @mo
function CC.SpawnDustRing(mo, speed, thrust, alwaysabove)
	local scale = mo.scale
	local momz = mo.momz
	if alwaysabove
		momz = -P_MobjFlip(mo)*abs($)
	end
	if abs(momz) < scale
		momz = $ < 0 and -scale or scale
	end
		
	local forwardangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	local sideangle = forwardangle + ANGLE_90
	local vangle = R_PointToAngle2(0, 0, FixedHypot(mo.momx, mo.momy), momz)
	
	local cosine = cos(vangle)
	local sine = sin(vangle)
	
	local radius = FixedDiv(mo.height, scale) >> 1
	local xspawn = -FixedMul(FixedMul(cos(forwardangle), cosine), radius)
	local yspawn = -FixedMul(FixedMul(sin(forwardangle), cosine), radius)
	local zoffset = -FixedMul(FixedMul(sine, radius), scale)
	
	local hthrust = 0
	local vthrust = 0
	if thrust
		hthrust = FixedMul(thrust, cosine)
		vthrust = FixedMul(thrust, sine)
	end
	
	cosine = FixedMul($, speed)
	sine = FixedMul($, speed)
	
	for i = 1, 16
		local dust = P_SpawnMobjFromMobj(mo, xspawn, yspawn, radius, mo.eflags & (MFE_UNDERWATER | MFE_TOUCHWATER) and MT_MEDIUMBUBBLE or MT_DUST)
		
		local a = i*ANGLE_22h + forwardangle
		
		local forwardthrust = FixedMul(cos(a), sine)
		local sidethrust = FixedMul(sin(a), speed)
		local zthrust = FixedMul(cosine, cos(a))
		
		dust.z = $ + zoffset
		
		P_Thrust(dust, forwardangle, forwardthrust - hthrust)
		P_Thrust(dust, sideangle, sidethrust)
		dust.momz = -zthrust - vthrust
		
		if dust.type == MT_DUST then dust.state = S_CHRISPY_DUST2 end
		dust.fuse = TICRATE/4
		dust.destscale = scale/8
		dust.scalespeed = FixedMul($, scale)
	end
end

// reduces @player's friction so they can gain speed from slopes
CC.Grease = function(player)
	local mo = player.mo
	local cmd = player.cmd
	
	if mo.friction < FRACUNIT
		mo.friction = $ + FixedMul(FRACUNIT - $, min(FRACUNIT, FixedHypot(cmd.forwardmove * FRACUNIT, cmd.sidemove * FRACUNIT) / 45))
	end
end

// returns whether @player can emote
CC.CanEmote = function(player)
	local mo = player.mo
	local chrispy = player.chrispy
	
	if not mo
	or not chrispy
		return false
	end
	
	return not (player.cmd.forwardmove or player.cmd.sidemove)
	and chrispy.emotewheel
	and not (chrispy.flags & (CF_BOUNCING|CF_DEFAULTDANCE))
	and (player.powers[pw_carry] == CR_NONE or player.powers[pw_carry] == CR_ROLLOUT)
	and player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE) ~= PF_JUMPED
	and not (player.pflags & (PF_SLIDING|PF_SPINNING))
	and not P_PlayerInPain(player)
	and player.playerstate == PST_LIVE
	and not (mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6)
	and mo.state ~= S_POINTY_FLIP
	and mo.state ~= S_CHRISPY_TRICK
	and mo.state ~= S_PLAY_GASP
	and player.panim ~= PA_EDGE
	and not player.actionstate // Battle
end

// returns whether @player can dab
CC.CanDab = function(player)
	local mo = player.mo
	local chrispy = player.chrispy
	
	if not mo
	or not chrispy
		return false
	end
	
	return not (player.cmd.forwardmove or player.cmd.sidemove)
	and not (chrispy.flags & (CF_CROUCHING|CF_BOUNCING|CF_DEFAULTDANCE))
	and (player.powers[pw_carry] == CR_NONE or player.powers[pw_carry] == CR_ROLLOUT)
	and player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE) ~= PF_JUMPED
	and not (player.pflags & (PF_SLIDING|PF_SPINNING))
	and not P_PlayerInPain(player)
	and player.playerstate == PST_LIVE
	and not (mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6)
	and mo.state ~= S_POINTY_FLIP
	and mo.state ~= S_CHRISPY_TRICK
	and mo.state ~= S_PLAY_GASP
	and not player.actionstate // Battle
end

// returns whether @player can use a Drop Dash (does not include button checks)
CC.CanUseDropDash = function(player)
	local mo = player.mo
	local shieldProperties = CC.GetShieldProperties(player.powers[pw_shield] & SH_NOSTACK)
	
	if srb2p // no dropdashing in SRB2 Persona
	or (player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE)) ~= PF_JUMPED // need to be in a damagejump
	or player.charability2 ~= CA2_SPINDASH // something else does not want us to spin
	or player.gotflag or player.gotcrystal // no abilities while holding the flag
	or mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6 // not while transforming
	or (player.pflags & PF_SHIELDABILITY and shieldProperties & CC.SHIELDFLAGS.NOABILITYCOMBO)
	or (not (player.charflags & SF_NOSHIELDABILITY) and shieldProperties & CC.SHIELDFLAGS.NOMULTIPLEDROPDASH)
		return false
	end
	return true
end

// returns the skin that @player should steal default stats from
function CC.GetStats(player)
	if player.bot == BOT_2PAI
		local leader = player.botleader
		if leader ~= nil
			return skins[leader.skin]
		end
	end
	
	return skins[player.skin]
end

// sets base movement stats to what is determined to be default
function CC.SetStats(player)
	local skin = CC.GetStats(player)
	
	player.normalspeed = skin.normalspeed
	player.acceleration = skin.acceleration
	player.accelstart = skin.accelstart
	player.thrustfactor = skin.thrustfactor
end

// returns this player's bot leader, built to account for 2.2.9 and 2.2.10
function CC.GetBotLeader(player)
	if BOT_2PAI // check for presence of !1383
		return player.botleader
	end
	return players[0]
end

// executes @function for each bot player following @leader
function CC.ExecuteForBots(leader, func)
	for player in players.iterate
		if player.bot == CC.BOT_2PAI
		and CC.GetBotLeader(player) == leader
			func(player, leader)
		end
	end
end

// swaps @player and @player2
function CC.SwapPlayers(player1, player2)
	local chrispy1 = player1.chrispy
	local chrispy2 = player2.chrispy
	local mo1 = player1.mo
	local mo2 = player2.mo
	
	local pos = {mo2.x, mo2.y, mo2.z}
	local drawangle = player2.drawangle
	local height1, height2 = player1.height, player2.height
	
	// color
	player1.skincolor, player2.skincolor = $2, $1
	mo1.color, mo2.color = $2, $1
	
	// skin
	mo1.skin, mo2.skin = $2, $1
	R_SetPlayerSkin(player1, mo1.skin)
	R_SetPlayerSkin(player2, mo2.skin)
	
	// chrispy tables
	for i, skin in ipairs(CC.SKINS)
		chrispy1[skin], chrispy2[skin] = $2, $1 // swaps skin-specific tables
	end
	mo1.fluffytails, mo2.fluffytails = $2, $1
	
	// retarget flickies
	for _, c in ipairs({chrispy1, chrispy2})
		local flickies = c.fluffy and c.fluffy.flickies
		if not flickies
			continue
		end
		for _, bird in ipairs(flickies)
			if bird.valid
				bird.tracer = c.mo
			end
		end
	end
	
	// powers
	for i, power in ipairs(CC.BOT_SWAPPOWERS)
		player1.powers[power], player2.powers[power] = $2, $1
	end
	
	// swap rollout rocks (because you can crouch on rollout rocks!)
	if player1.powers[pw_carry] == CR_ROLLOUT
	or player2.powers[pw_carry] == CR_ROLLOUT
		player1.powers[pw_carry], player2.powers[pw_carry] = $2, $1
		mo1.tracer, mo2.tracer = $2, $1
	end
	
	// drawangle
	if player2.pflags & PF_DIRECTIONCHAR
		player2.drawangle = player1.drawangle
	end
	if player1.pflags & PF_DIRECTIONCHAR
		player1.drawangle = drawangle
	end
	
	// height
	player1.height = height2
	player2.height = height1
	
	// position
	P_TeleportMove(mo2, mo1.x, mo1.y, mo1.z)
	P_TeleportMove(mo1, unpack(pos))
	
	// reset shield visuals
	P_SpawnShieldOrb(player1)
	P_SpawnShieldOrb(player2)
end

// returns whether this player uses special super animations (e.g. Pointy freeflight)
CC.UsingSuperAnims = function(player)
	return player.powers[pw_super] and player.chrispy and player.chrispy.superanims
end

// returns whether this player is in vanilla SRB2's hyper form
CC.GodModeHyper = function(player)
	return player.pflags & PF_GODMODE and player.powers[pw_super]
end

// returns whether this player is in xmom's hyper form
CC.XMomHyper = function(player)
	return player.hyper
	and player.powers[pw_super]
	and player.hyper.capable
	and player.hyper.isunlocked
end

// returns whether this player is in any hyper form
CC.IsHyper = function(player)
	return CC.GodModeHyper(player) or CC.XMomHyper(player)
end

// if @player is hyper, returns which color the player should be
// otherwise, returns SKINCOLOR_NONE
CC.GetHyperColor = function(player)
	if CC.XMomHyper(player)
		return (SKINCOLOR_SUPER_HYPER_1 + abs( ( (player.powers[pw_super] >> 1) % 9) - 4)) // color calculation from xmom 1.2
	end
	if CC.GodModeHyper(player)
		return SKINCOLOR_HYPERPOINTYSILVER + (leveltime >> 1) % 7 + (skins[player.skin].supercolor - SKINCOLOR_SUPERLIGHTBULB1)
	end
	return SKINCOLOR_NONE
end