/*
ChrispyChars
	by Lach

v1.2
*/

/*
COMPATIBILITY NOTES:
	- as of v1.2, compatibility notes from v1 and v1.1 are no longer necessary!
	- see CHRISPYCHARS-DEFS and CHRISPYCHARS-FUNCS for info on shield stuff
*/

local CC = ChrispyChars

CC.SKINS = {"pointy", "fluffy"}

local DROPDASH_MINCHARGE = 25*FRACUNIT // minimum drop dash charge
local DROPDASH_MAXCHARGE = 70*FRACUNIT // maximum drop dash charge
local DROPDASH_CHARGESPEED = FRACUNIT // drop dash charge to add per tic
local DROPDASH_DELAY = 8 // number of frames to charge drop dash

local DABTRANS_TIME = 8*TICRATE
local DABTRANS_THRESHOLD = 3*TICRATE
local DABTRANS_COLORINTERVAL = DABTRANS_THRESHOLD/16

states[S_CHRISPY_DABTRANS].tics = DABTRANS_THRESHOLD

local DAB_DISTANCE = 512*FRACUNIT // distance from dabbing player where screenshake begins to take effect
local DAB_POWER = 6*FRACUNIT // screenshake strength at minimum distance

rawset(_G, "DROPDASH_DELAY", DROPDASH_DELAY)

local BUTTONS = {BT_JUMP, BT_USE, BT_TOSSFLAG, BT_CUSTOM1, BT_CUSTOM2, BT_CUSTOM3,
	BT_ATTACK, BT_FIRENORMAL, BT_WEAPONNEXT, BT_WEAPONPREV} // extra buttons, for action reassignment

local flagnum = 0
rawset(_G, "CC_RegisterFlag", function(name)
	assert(flagnum < 32, "Hit flag cap, could not assign flag "..name.."!")
	rawset(_G, name, 1 << flagnum)
	flagnum = $ + 1
	return _G[name]
end)

local FLAGS = {
	CC_RegisterFlag("CF_DROPDASHCHARGE"),
	CC_RegisterFlag("CF_USEDDROPDASH"),
	CC_RegisterFlag("CF_FASTROLL"),
	CC_RegisterFlag("CF_CROUCHING"),
	CC_RegisterFlag("CF_DABBING"),
	CC_RegisterFlag("CF_EMOTEWHEEL"),
	CC_RegisterFlag("CF_NIGHTSRISE"),
	CC_RegisterFlag("CF_SUPERBOT")
}
rawset(_G, "CF_SHAREDMASK", FLAGS[#FLAGS]|(FLAGS[#FLAGS] - 1) - (FLAGS[1] - 1))
rawset(_G, "CF_DROPDASHMASK", CF_DROPDASHCHARGE|CF_USEDDROPDASH)
FLAGS = nil

if not CC_TrickSkins
	rawset(_G, "CC_TrickSkins", {})
end
CC_TrickSkins["tails"] = { // MT_TAILSOVERLAY requires a different default, and this is pretty much the only sane option...
	{
		sprite2 = SPR2_SWIM,
		rotation = "random",
		tics = 3,
		frames = 4
	}
}
CC_TrickSkins["fang"] = { // a more solid example to follow
	{
		sprite2 = SPR2_SPNG,
		rotation = 1,
		tics = 0
	},
	{
		sprite2 = SPR2_EDGE,
		rotation = -1,
		tics = 2,
		frames = 4
	},
	{
		sprite2 = SPR2_LAND,
		rotation = 0,
		tics = 3,
		frames = 3,
		frame = FF_VERTICALFLIP
	}
}

rawset(_G, "CC_DEFAULT_CONFETTI_COLORS", { // overwriting the list from CC_TRICKS!
	SKINCOLOR_NONE,
	SKINCOLOR_NONE,
	SKINCOLOR_SUPERSTREETLAMP1,
	SKINCOLOR_CARDINAL,
	SKINCOLOR_SUPERLIGHTBULB5,
	SKINCOLOR_BEEG,
	SKINCOLOR_HYPERFLUFFYSKY,
	SKINCOLOR_TAFFY,
	SKINCOLOR_DREAMY,
})

rawset(_G, "CC_Footsteps", {})

rawset(_G, "FSF_NOTRANSITION", 1 << 0)
rawset(_G, "FSF_MOVEMENTONLY", 1 << 1) // only affects repeatsound
rawset(_G, "FSF_AERIAL", 1 << 2)

rawset(_G, "CC_GetCVar", function(var)
	return var and var.value or 1
end)
rawset(_G, "CC_Renderer", CV_FindVar("renderer"))

local function valid(mo)
	return mo and mo.valid
end

local function SetFrame(mo, framenum)
	local currentframe = mo.frame & FF_FRAMEMASK
	if currentframe ~= framenum
		mo.frame = framenum | ($ & ~FF_FRAMEMASK)
	end
end

local spr2numframes = {}
local function GetSprite2NumFrames(skin, spr2)
	local frame
	
	if not spr2numframes[skin]
		spr2numframes[skin] = {}
	else
		frame = spr2numframes[skin][spr2]
	end
	
	if frame ~= nil
		return frame
	end
	
	local statenum = S_UNKNOWN
	local state = states[statenum]
	local statedef = {
		state.sprite,
		state.frame,
		state.tics,
		state.action,
		state.var1,
		state.var2,
		state.nextstate
	}
	local thok = P_SpawnMobj(0, 0, 0, MT_THOK)
	state.sprite = SPR_PLAY
	state.frame = spr2
	state.action = nil
	thok.skin = skin
	thok.state = statenum
	frame = -1
	while thok.frame & FF_FRAMEMASK > frame
		frame = thok.frame & FF_FRAMEMASK
		thok.state = statenum
	end
	frame = $ + 1
	spr2numframes[skin][spr2] = frame
	states[statenum] = statedef
	P_RemoveMobj(thok)
	return frame
end

local function AngleStand(player) // directionchar hack
	player.skidtime = max($, 1)
end

function CC.StartHitLag(player, tics)
	player.chrispy.hitlag = max($, tics or 0)
	player.mo.flags = $ | MF_NOTHINK
end

function CC.UpdateHitLag(player)
	local chrispy = player.chrispy
	
	if chrispy.hitlag
		chrispy.hitlag = $ - 1
		if chrispy.hitlag == 0
			player.mo.flags = $ & ~MF_NOTHINK
			return
		end
		
		player.pflags = $ | PF_FULLSTASIS
	end
end

local function Undab(player)
	local mo = player.mo
	local chrispy = player.chrispy
	
	if mo.state == S_CHRISPY_DAB
	or mo.state == S_CHRISPY_HATER
		if P_IsObjectOnGround(mo) or player.powers[pw_carry] == CR_ROLLOUT
			if player.speed or player.cmd.forwardmove or player.cmd.sidemove
				mo.state = S_PLAY_WALK
			else
				mo.state = S_PLAY_STND
			end
			AngleStand(player)
		else
			if player.pflags & PF_AUTOBRAKE
			and mo.state == S_CHRISPY_DAB
				player.pflags = $ | PF_APPLYAUTOBRAKE
			end
			mo.state = S_PLAY_FALL
		end
	end
	
	if chrispy.supertrans <= DABTRANS_THRESHOLD
	and not player.powers[pw_super]
		S_StartSound(mo, sfx_rscrch)
		if mo.color >= SKINCOLOR_CHRISPYTRANS1
		and mo.color < SKINCOLOR_CHRISPYTRANS1 + CC_TRANSCOLOR_SLOTS
			mo.color = player.powers[pw_dye] or player.skincolor
		end
		
		if valid(chrispy.transoverlay)
			P_RemoveMobj(chrispy.transoverlay)
			chrispy.transoverlay = nil
		end
	end
	player.chrispy.flags = $ & ~CF_DABBING
end

local function GetWheelEmote(chrispy)
	local emote = chrispy.emote
	if not valid(emote)
		local mo = chrispy.mo
		emote = P_SpawnMobj(mo.x, mo.y, mo.z, MT_EMOTE)
		emote.tracer = mo
		emote.target = mo
		emote.customskin = "pointy"
		chrispy.emote = emote
	end
	return emote
end

local slot = 0
local function DoSuperTransColor(player, num)
	local mo = player.mo
	local skin = mo.skin
	
	if num > 15
		mo.color = player.powers[pw_dye] or player.skincolor
	elseif num < -15
		mo.color = CC.GetHyperColor(player)
			or skins[skin].supercolor + min(4, abs(num) - 16 + 2)
	else
		mo.color = SKINCOLOR_CHRISPYTRANS1 + slot
		mo.frame = $ | FF_FULLBRIGHT
		
		slot = ($ + 1) % CC_TRANSCOLOR_SLOTS
		
		// I'm not going to freeslot a transformation color for all 32 players, so basically if the displayplayer is transforming then we prioritize them, otherwise I don't give a fuck
		if displayplayer ~= player
			local dmo = displayplayer.mo
			if dmo and dmo.color == mo.color
			and mo.state >= S_PLAY_SUPER_TRANS1
			and mo.state <= S_PLAY_SUPER_TRANS6
			and #displayplayer < #player
				return
			end
		end
		
		local index = 0
		local index2 = 0
		local color = num > 0 and (player.powers[pw_dye] or player.skincolor)
			or CC.GetHyperColor(player)
			or skins[skin].supercolor + 2 // (middle supercolor)
		
		num = abs(num)
		
		for i = 1, 16 - num
			skincolors[mo.color].ramp[index] = 0
			index = $ + 1
		end
		while index < 16
			skincolors[mo.color].ramp[index] = skincolors[color].ramp[index2]
			index = $ + 1
			index2 = $ + 1
		end
	end
end

local function CheckShieldAbility(player, grounded)
	local chrispy = player.chrispy
	
	if grounded
		player.charflags = $ | SF_NOSHIELDABILITY
		return
	end
	
	if chrispy.flags & CF_USEDDROPDASH
	or not (player.pflags & PF_JUMPED)
		player.charflags = $ & ~SF_NOSHIELDABILITY
		return
	end
end

rawset(_G, "IsChrispySkin", function(skin)
	for i = 1, #CC.SKINS
		if skin == CC.SKINS[i]
		or skin == #skins[CC.SKINS[i]]
			return true
		end
	end
	return false
end)

rawset(_G, "CC_Autobrake", function(player, factor) // autobrake logic copied from source
	local mo = player.mo
	local speed = FixedHypot(mo.momx - player.cmomx, mo.momy - player.cmomy)
	local acceleration = (player.accelstart + (FixedDiv(speed, mo.scale) >> FRACBITS) * player.acceleration) * (player.thrustfactor or skins[mo.skin].thrustfactor) * 20
	local moveangle = R_PointToAngle2(0, 0, mo.momx - player.cmomx, mo.momy - player.cmomy)
	
	if not factor
		factor = P_IsObjectOnGround(mo) and FRACUNIT or (FRACUNIT >> 1)
	end
	
	acceleration = FixedMul($, factor)
	
	if mo.movefactor ~= FRACUNIT
		acceleration = FixedMul($ << FRACBITS, mo.movefactor) >> FRACBITS
	end
	
	P_Thrust(mo, moveangle, -min(acceleration, speed))
end)

local function SpawnCorona(x, y, z, scale, flip)
	local mo = P_SpawnMobj(x, y, z, MT_CHRISPY_CORONA)
	local beam = mo
	mo.scale = scale
	mo.blendmode = AST_ADD
	
	if flip
		mo.eflags = $ | MFE_VERTICALFLIP
		mo.z = $ - mo.height
	end
	
	// spawn beams
	for i = 1, 6
		beam.tracer = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_OVERLAY)
		beam = $.tracer
		
		beam.target = mo
		beam.blendmode = AST_ADD
		beam.state = mo.state
		beam.frame = $ & ~FF_FRAMEMASK | (B + P_RandomKey(8))
		beam.rollangle = FixedAngle(FixedMul(P_RandomFixed(), 360*FRACUNIT))
		beam.movedir = (-1 + P_RandomKey(2) * 2) * ANG1
		beam.extravalue1 = (-1 + P_RandomKey(2) * 2)
	end
	return mo
end

local function StartHighFive(player1, player2)
	for i = 1, 2
		local chrispy = player1.chrispy
		
		chrispy.highFivePartner = player2
		chrispy.highFiveState = 1
		chrispy.highFiveDelay = CC.HIGHFIVE_DELAY
		
		S_StartSound(nil, sfx_cdfm60, player1)
		
		player1, player2 = $2, $1
	end
end

local function HighFiveContact(player, player2)
	local mo, mo2 = player.mo, player2.mo
	local scale = (mo.scale + mo2.scale)/2
	local x, y = mo.x/2 + mo2.x/2, mo.y/2 + mo2.y/2
	
	if mo.eflags & MFE_VERTICALFLIP
		local z = (mo.z + mo.height)/2 + (mo2.z + mo2.height)/2
		S_StartSound(SpawnCorona(x, y, z, scale, true), sfx_hifive)
	else
		local z = (mo.z/2 + mo2.z/2)
		S_StartSound(SpawnCorona(x, y, z, scale), sfx_hifive)
	end
	
	for i = 1, 2
		local mo = player.mo
		
		CC.StartHitLag(player, TICRATE/2)
		mo.tics = 8
		player.chrispy.highFiveState = 3
		
		player = player2
	end
end

function CC.ActiveHighFiveState(player)
	return player.chrispy.highFiveState > 0 and player.chrispy.highFiveState < 3
end

local function HandleHighFive(player)
	local mo = player.mo
	local chrispy = player.chrispy
	
	// searching for a high five partner
	if chrispy.highFiveState == 0
		local closestDist = FixedMul(CC.HIGHFIVE_RADIUS, mo.scale)
		local closestPlayer
		local closestChrispy
		
		for player2 in players.iterate
			local mo2 = player2.mo
			local chrispy2 = player2.chrispy
			
			if player == player2
			or not chrispy2
			or chrispy2.highFiveState ~= 0
			or not mo2
			or mo2.state ~= S_CHRISPY_HIGHFIVE
			or (mo2.eflags & MFE_VERTICALFLIP ~= mo.eflags & MFE_VERTICALFLIP)
				continue
			end
			
			local dist = FixedHypot(FixedHypot(mo2.x - mo.x, mo2.y - mo.y),
				(mo2.z + (mo2.height >> 1)) - (mo.z + (mo.height >> 1)))
			
			if dist > closestDist
				continue
			end
			
			closestPlayer = player2
			closestChrispy = chrispy2
			closestDist = dist
		end
		
		if not closestPlayer
			return
		end
		
		StartHighFive(player, closestPlayer)
		if #closestPlayer < #player
			return // ensure both players' timers start counting down on the same tic
		end
	end
	
	// performing a high five
	if CC.ActiveHighFiveState(player)
		local partner = chrispy.highFivePartner
		local partnerMo = valid(partner) and partner.mo
		
		// if the partner exits the commitment, return to search state
		if not valid(partner)
		or not partner.chrispy
		or not partnerMo
		or partnerMo.state ~= S_CHRISPY_HIGHFIVE
		or (partnerMo.eflags & MFE_VERTICALFLIP ~= mo.eflags & MFE_VERTICALFLIP)
			if chrispy.highFiveState > 1
				mo.state = S_PLAY_FALL
			end
			chrispy.highFiveState = 0
			return
		end
		
		// initiation: pause for a second so you see your partner
		if chrispy.highFiveState == 1
			// face your partner & force the camera to swing to a side view
			// this doesn't work in old analog mode but I don't care, all my homies hate old analog mode
			player.drawangle = R_PointToAngle2(mo.x, mo.y, partnerMo.x, partnerMo.y)
			mo.angle = player.drawangle + ANGLE_90
			
			// once the timer is up, leap towards your partner
			chrispy.highFiveDelay = $ - 1
			if chrispy.highFiveDelay <= 0
				local flip = P_MobjFlip(mo)
				local grav = P_GetMobjGravity(mo)
				local scale = (mo.destscale + partnerMo.destscale)/2
				local hSpeed = FixedMul(CC.HIGHFIVE_SPEED, scale)
				local hDist = FixedHypot(mo.x - partnerMo.x, mo.y - partnerMo.y)/2 - (mo.radius + partnerMo.radius)
				local goalZ, vDist
				
				if mo.eflags & MFE_VERTICALFLIP
					goalZ = min(mo.z + mo.height, partnerMo.z + partnerMo.height) - FixedMul(CC.HIGHFIVE_HEIGHT, scale)
					vDist = goalZ - (mo.z + mo.height)
				else
					goalZ = max(mo.z, partnerMo.z) + FixedMul(CC.HIGHFIVE_HEIGHT, scale)
					vDist = goalZ - mo.z
				end
				
				mo.z = $ + flip // raise the player above the ground
				if mo.standingslope // this will trigger a slope launch next tic, so we have to do ugly hacks to bypass this...
					local x, y, z = mo.x, mo.y, mo.z // store their current position
					P_XYMovement(mo) // trigger the slope launch early (AWFUL)
					P_TeleportMove(mo, x, y, z) // teleport them back to their initial position
				end
				mo.momz = flip*FixedSqrt(abs(2 * FixedMul(grav, vDist)))
				P_InstaThrust(mo, player.drawangle, -FixedDiv(hDist, FixedDiv(mo.momz, grav)))
				
				mo.state = $
				player.panim = PA_FALL
				chrispy.highFiveState = 2
				// I don't know why but the sounds stop immediately if they just play
				if player == displayplayer
					S_StartSound(nil, sfx_ngjump, player)
				elseif partner != displayplayer
					S_StartSound(mo, sfx_ngjump)
				end
			end
		end
		
		// during the leap
		if chrispy.highFiveState == 2
			// check whether we made contact with our partner!
			local scale = (mo.destscale + partnerMo.destscale)/2
			local zDist
			if mo.eflags & MFE_VERTICALFLIP
				zDist = abs((mo.z + mo.height) - (partnerMo.z + partnerMo.height))
			else
				zDist = abs(mo.z - partnerMo.z)
			end
			
			if partner.chrispy.highFiveState == chrispy.highFiveState // no highfiving people on the ground LOL
			and zDist <= FixedMul(abs(CC.HIGHFIVE_PALMHEIGHT), scale)
				local dist = FixedDiv(FixedHypot(mo.x - partnerMo.x, mo.y - partnerMo.y), scale)
				if dist <= CC.HIGHFIVE_CONTACT
				and dist >= CC.HIGHFIVE_CONTACT - CC.HIGHFIVE_PALMWIDTH
					HighFiveContact(player, partner)
				end
			end
					
			// disable controls
			player.pflags = $ | PF_FULLSTASIS
		end
	end
end

// R Emeralds

local emmies = {}
local birbs = {}

local function REmeraldOverlay(mo)
	if mo.cc_overlay
		return
	end
	
	local frame = mo.frame & FF_FRAMEMASK
	local overlay = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_OVERLAY)
	overlay.target = mo
	overlay.state = S_CHRISPY_REMERALD_BACK
	overlay.angle = mo.angle
	if frame == 2 // yellow emerald
	or frame == 5 // red emerald
		overlay.frame = $ & ~FF_TRANSMASK | FF_TRANS80
	elseif frame == 4 // orange emerald
		overlay.frame = $ & ~FF_TRANSMASK | FF_TRANS90
	end
	mo.cc_overlay = overlay
end

local function REmeraldChange(mo)
	REmeraldOverlay(mo)
	mo.sprite = SPR_REMG
	mo.frame = $ | FF_TRANS10
end

local function AllPlayersChrispy()
	local ret = false
	
	for player in players.iterate
		local mo = player.mo
		if mo
			ret = IsChrispySkin(mo.skin)
			if not ret // there is a non-Chrispychar in play, so don't change the emeralds
				return false
			end
		end
	end
	
	return ret
end

local function REmeraldSpawn(mo)
	if not leveltime // map loading, no players are spawned yet
		return
	end
	
	if AllPlayersChrispy()
		REmeraldChange(mo)
	end
end

local function REmeraldDie(mo)
	if valid(mo.cc_overlay)
		P_RemoveMobj(mo.cc_overlay)
	end
end

local function REmeraldSpin(mo)
	if mo.sprite == SPR_REMG
		mo.angle = $ - ANG10
	end
end

for i = 0, 6
	CC.AddHook("MobjSpawn", REmeraldSpawn, MT_EMERALD1 + i)
	CC.AddHook("MobjDeath", REmeraldDie, MT_EMERALD1 + i)
	CC.AddHook("MobjThinker", REmeraldSpin, MT_EMERALD1 + i)
end

CC.AddHook("MapLoad", do
	if AllPlayersChrispy()
		local starttype = mobjinfo[MT_EMERALD1].doomednum
		local endtype = mobjinfo[MT_EMERALD7].doomednum
		for mt in mapthings.iterate
			if mt.type >= starttype
			and mt.type <= endtype
			and valid(mt.mobj)
				REmeraldChange(mt.mobj)
			end
		end
	end
end)

CC.AddHook("MobjSpawn", function(mo)
	if G_IsSpecialStage(gamemap)
		table.insert(emmies, mo)
	end
end, MT_GOTEMERALD)

CC.AddHook("MobjSpawn", function(mo)
	local n
	if gamemap >= sstage_start
	and gamemap <= sstage_end
		local map = gamemap - sstage_start + 1
		if map == 1 // green
			n = 4
		elseif map == 2 // pink
			n = 2
		elseif map == 6 // magenta
			n = 3
		elseif map == 4 // cyan, closest to blue
			n = 1
		end
	end
	n = $ or P_RandomKey(#CC_FLICKY_SPRITES) + 1
	table.insert(birbs, mo)
	mo.cc_sprite = CC_FLICKY_SPRITES[n]
	mo.cc_color = CC_FLICKY_COLORS[n]
end, MT_NIGHTOPIANHELPER)

CC.AddHook("ThinkFrame", do
	for i = #emmies, 1, -1
		local mo = emmies[i]
		if not valid(mo)
			table.remove(emmies, i)
			continue
		end
		
		local target = mo.target
		
		if not (valid(target) and IsChrispySkin(target.skin))
			REmeraldDie(mo)
			table.remove(emmies, i)
			continue
		end
		
		REmeraldChange(mo)
		mo.angle = -leveltime*ANG10
	end
	
	for i = #birbs, 1, -1
		local mo = birbs[i]
		
		local target = valid(mo) and mo.target
		
		if not (valid(target) and IsChrispySkin(target.skin))
			table.remove(birbs, i)
			continue
		end
		
		mo.sprite = mo.cc_sprite
		mo.color = mo.cc_color
	end
end)

// fan sector IDs
local fansectors = {}
CC.AddHook("MapChange", do
	fansectors = {}
end)

// main ThinkFrames

CC.AddHook("PreThinkFrame", do // oh my god fuck SRB2
	for player in players.iterate
		local chrispy = player.mo and player.chrispy
		
		if not chrispy then continue end
		
		if chrispy.flags & CF_EMOTEWHEEL
			local emotewheel = chrispy.emotewheel
			emotewheel.left = player.cmd.sidemove < 0 and $ + 1 or 0
			emotewheel.right = player.cmd.sidemove > 0 and $ + 1 or 0
			player.cmd.sidemove = 0
		end
	end
end)		

CC.AddHook("ThinkFrame", do
	local dabshake = 0 // for dab screenshake, see end of function
	local dabbers = {}
	local haters = {}
	
	for player in players.iterate
		local mo = player.mo
		
		if not mo
			continue
		end
		if not IsChrispySkin(mo.skin)
			local chrispy = player.chrispy
			if chrispy
				local fluffy = chrispy.fluffy
				
				if mo.state >= CC.FIRSTSTATE
				and mo.state <= CC.LASTSTATE
					mo.state = S_PLAY_STND
				end
				
				if chrispy.flags & CF_SUPERPOINTY
					mo.flags = $ & ~MF_NOGRAVITY
				end
				
				if chrispy.hitlag
					mo.flags = $ & ~MF_NOTHINK
				end
				
				if fluffy
				and valid(fluffy.gauge)
					P_RemoveMobj(fluffy.gauge)
				end
				
				player.chrispy = nil
				mo.chrispy = nil
			end
			continue
		end
		
		local chrispy = player.chrispy
		
		if not chrispy
		or chrispy.mo ~= mo
			if chrispy
			and chrispy.transfershield
				player.powers[pw_shield] = chrispy.transfershield
				P_SpawnShieldOrb(player)
			end
			
			player.chrispy = {}
			local chrispy = player.chrispy
			mo.chrispy = chrispy
			chrispy.mo = mo
			chrispy.player = player
			
			chrispy.time = -1
			chrispy.flags = 0
			chrispy.deathflags = 0
			chrispy.dropdash = 0
			chrispy.ddtrans = 0
			chrispy.prevstep = 0
			chrispy.spinparticles = {}
			chrispy.fantics = 0
			chrispy.emotewheel = {
				scale = 0,
				selection = 1,
				angle = 0,
				left = 0,
				right = 0,
				duration = -1,
				tics = -1
			}
			chrispy.prevmomentum = {mo.momx, mo.momy, mo.momz}
			chrispy.supertrans = DABTRANS_TIME
			chrispy.nightsframe = 0
			chrispy.hitlag = 0
			chrispy.highFiveState = 0
			chrispy.highFiveDelay = 0
			chrispy.lastSwappedPlayerNum = 0
			
			chrispy.buttons = {}
			
			for i = 1, #BUTTONS
				chrispy.buttons[BUTTONS[i]] = 0
			end
			
			player.height = skins[mo.skin].height
		end
		
		local chrispy = player.chrispy
		local flip = P_MobjFlip(mo)
		local actuallygrounded = P_IsObjectOnGround(mo)
		local grounded = actuallygrounded or player.powers[pw_carry] == CR_ROLLOUT
		local looking = mo.state == S_PLAY_LOOK
		local dabbutton = player.cc_buttonmap and player.cc_buttonmap.dab or CC_DEFAULTBUTTONS.dab
		local crouchbutton = player.cc_buttonmap and player.cc_buttonmap.crouch or CC_DEFAULTBUTTONS.crouch
		local dancebutton = player.cc_buttonmap and player.cc_buttonmap.dance or CC_DEFAULTBUTTONS.dance
		local emotewheel = chrispy.emotewheel
		local activeemote = false
		
		chrispy.time = max($ + 1, 0)
		
		// count buttons held
		
		for i = 1, #BUTTONS
			chrispy.buttons[BUTTONS[i]] = player.cmd.buttons & BUTTONS[i] and $ + 1 or 0
		end
		
		// NiGHTS!
		if player.powers[pw_carry] == CR_NIGHTSMODE
			local feet = chrispy.nightsfeet
			local angle
			local addangle = 0
			local invert = 0
			local factor = 6
			local frame
			
			// spawn feet in if not already
			if not valid(feet)
				feet = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_OVERLAY)
				feet.state = S_CHRISPY_NIGHTSFEET
				feet.target = mo
				feet.tracer = feet.target
				chrispy.nightsroll = mo.rollangle
				chrispy.nightsfeet = feet
			end
			
			// calculate whether flying left or right
			if splitscreen
				angle = (player.cmd.angleturn << 16) - mo.angle
			else
				angle = R_PointToAngle(mo.x, mo.y) - mo.angle
			end
			
			if (player.flyangle >= 90 and player.flyangle <= 270)
				if not (player.flyangle == 270 and (angle >= 0))
				and not (player.flyangle == 90 and (angle >= 0))
					invert = 1
					addangle = ANGLE_180
				end
			end
			
			// transforming
			if mo.state >= S_PLAY_NIGHTS_TRANS1
			and mo.state <= S_PLAY_NIGHTS_TRANS6
				if mo.state == S_PLAY_NIGHTS_TRANS5
				and mo.tics == 1 // skip ghost spawning
					P_SetMobjStateNF(mo, states[mo.state].nextstate)
					mo.tics = $ + 1
				end
				frame = A + ((mo.state - S_PLAY_NIGHTS_TRANS1 + 1)/2) & 1
				
			// standing/floating
			elseif mo.state == S_PLAY_NIGHTS_STAND
			or mo.state == S_PLAY_NIGHTS_FLOAT
				if player.exiting // exiting animation
					if mo.tics >= 0
						chrispy.nightsframe = 0
					elseif leveltime & 1
						chrispy.nightsframe = ($ + 1) % 4
					end
					frame = I + chrispy.nightsframe
					player.drawangle = mo.angle
					mo.rollangle = angle < 0 and -ANGLE_90 or ANGLE_90
				else
					frame = C
				end
				mo.tics = -1
			
			// capsule pull
			elseif mo.state == S_PLAY_NIGHTS_PULL
				frame = H
				mo.tics = -1
			
			// drill
			elseif mo.state == S_PLAY_NIGHTS_DRILL
			or mo.state == S_PLAY_NIGHTS_ATTACK
				if mo.tics >= 0
					chrispy.nightsframe = 0
					mo.tics = -1
				elseif leveltime & 1
					chrispy.nightsframe = ($ + 1) % 4
				end
				frame = I + chrispy.nightsframe
			end
			
			// flying/stunned
			if mo.state == S_PLAY_NIGHTS_FLY
			or mo.state == S_PLAY_NIGHTS_STUN				
				if chrispy.nightsroll - addangle < mo.rollangle
					chrispy.flags = ($ | CF_NIGHTSRISE) ^^ invert*CF_NIGHTSRISE
				else
					chrispy.flags = ($ & ~CF_NIGHTSRISE) ^^ invert*CF_NIGHTSRISE
				end
				
				chrispy.nightsroll = $ + (mo.rollangle + addangle - chrispy.nightsroll) / factor
				
				if mo.state == S_PLAY_NIGHTS_STUN
					mo.rollangle = $ + (invert or -1)*ANG20
					factor = 4
					frame = F
				elseif chrispy.flags & CF_NIGHTSRISE
				and abs(chrispy.nightsroll - mo.rollangle - addangle) > 5*ANG1
					frame = F
				else
					frame = D
				end
				
				mo.tics = -1 // no need to calculate any animation based stuff if we're not animating
				feet.flags2 = $ & ~MF2_DONTDRAW | (mo.flags2 & MF2_DONTDRAW)
				feet.color = mo.color
				feet.frame = frame + 1
				feet.rollangle = chrispy.nightsroll + addangle
			else
				feet.flags2 = $ | MF2_DONTDRAW
				if frame <= C
					chrispy.nightsroll = ANGLE_90
				else
					chrispy.nightsroll = mo.rollangle + addangle
				end
			end
			
			// set color
			if not multiplayer
			and not player.powers[pw_dye]
				mo.color = SKINCOLOR_DREAMY
			end
			
			// set sprite & frame
			if frame ~= nil
				mo.sprite = SPR_NIGH
				mo.frame = frame | ($ & ~FF_FRAMEMASK)
			end
			
			continue // no need to run any extra gamelogic
		elseif valid(chrispy.nightsfeet)
			P_RemoveMobj(chrispy.nightsfeet)
			chrispy.nightsfeet = nil
		end
		
		// ChrispyChars bots can copy some status data from the main player
		
		if player.bot ~= CC.BOT_NONE
			CC.UpdateBot(player)
		end
		
		// don't let Fluffy carry you if you hit the ground
		
		if grounded and not chrispy.grounded
			CC.ExecuteForBots(player, function(bot)
				if player.powers[pw_carry] == CR_NONE
					bot.pflags = $ & ~PF_CANCARRY
				end
			end)
		end
		
		// hitlag
		
		CC.UpdateHitLag(player)
		
		// Death effects
		
		if player.playerstate == PST_DEAD
			// if the player just died, they might have a deathtype
			local deathtype = chrispy.deathtype
			if deathtype
				if deathtype == DMG_ELECTRIC
					chrispy.deathflags = $ | 1 | 2
					S_StartSound(mo, sfx_buzz2)
				elseif deathtype == DMG_FIRE
					local overlay = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_OVERLAY)
					overlay.target = mo
					overlay.skin = mo.skin
					overlay.state = mo.state
					overlay.frame = $ | FF_TRANS90
					overlay.color = SKINCOLOR_JET
					overlay.colorized = true
					chrispy.deathflags = $ | 1 | 4
					chrispy.deathoverlay = overlay
				elseif deathtype == DMG_SPIKE
					for i = 1, 3
						mo.state = S_CHRISPY_DEATH
					end
				elseif deathtype == DMG_CRUSHED
					for i = 1, 4
						mo.state = S_CHRISPY_DEATH
					end
					mo.height = 0 // flatten the player!
					mo.flags = $ & ~(MF_NOGRAVITY|MF_NOCLIPHEIGHT)
					mo.momz = 0
					chrispy.deathflags = $ | 8
				end
				chrispy.deathtype = 0
			end
			
			local insector = mo.z >= mo.floorz and mo.z + mo.height <= mo.ceilingz
			local particlestates = {}
			
			// smoke
			if chrispy.deathflags & 1
				table.insert(particlestates, mo.eflags & MFE_UNDERWATER and S_MEDIUMBUBBLE or S_SMOKE1)
			end
			
			// electrocution
			if chrispy.deathflags & 2
				local n = chrispy.time % 8
				
				if not (n & 1)
					n = $ / 2
					if n == 0
						mo.state = S_PLAY_DEAD
						mo.color = player.powers[pw_dye] or player.skincolor
						mo.colorized = false
					elseif n & 1
						mo.state = S_CHRISPY_DEATH
						mo.frame = FF_FULLBRIGHT
						mo.color = SKINCOLOR_CHRISP
						mo.colorized = true
					else
						mo.state = S_CHRISPY_DEATH
						mo.frame = $ | FF_FULLBRIGHT
						mo.color = player.powers[pw_dye] or player.skincolor
						mo.colorized = false
					end
				end
				
				table.insert(particlestates, S_WATERZAP)
			end
			
			// burnt
			if chrispy.deathflags & 4
				local overlay = chrispy.deathoverlay
				if overlay.valid
					if chrispy.time & 1
						overlay.frame = $ - FF_TRANS10
					end
					if not (overlay.frame & FF_TRANSMASK)
						mo.color = overlay.color
						mo.colorized = overlay.colorized
						if overlay.color ~= SKINCOLOR_CHRISP
							overlay.color = SKINCOLOR_CHRISP
							overlay.frame = $ | FF_TRANS90
						else
							P_RemoveMobj(overlay)
						end
					end
				end
				table.insert(particlestates, mo.eflags & MFE_UNDERWATER and S_TRAINDUST or S_SPINDUST_FIRE1)
			end
			
			// crushed
			if chrispy.deathflags & 8
				if mo.eflags & MFE_VERTICALFLIP
					mo.z = mo.ceilingz
				else
					mo.z = mo.floorz
				end
			end
			
			// drown
			if mo.state == S_PLAY_DRWN
			and mo.eflags & MFE_UNDERWATER
				mo.momz = $ + flip*(mo.scale/3)
				
				if insector
				and P_RandomChance(FRACUNIT >> 1)
					P_SpawnMobjFromMobj(mo, 0, 0, 3*player.height >> 2,
						P_RandomChance(FRACUNIT/3) and MT_MEDIUMBUBBLE or MT_SMALLBUBBLE)
				end
			end
			
			// spawn particles
			if #particlestates
			and insector
				local angle = P_RandomFixed() * FRACUNIT
				local r = mo.radius
				
				local part = P_SpawnMobjFromMobj(mo,
					P_ReturnThrustX(mo, angle, r),
					P_ReturnThrustY(mo, angle, r),
					FixedMul(P_RandomFixed(), mo.height),
					MT_PARTICLE)
				part.state = particlestates[P_RandomKey(#particlestates) + 1]
				part.destscale = 0
				part.scalespeed = part.scale / 24
				part.fuse = 24
			end
		end
		
		// Super transformation
		
		if mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6
		or CC.XMomHyper(player) and mo.state >= S_PLAY_HYPERTRNS_1 and mo.state <= S_PLAY_HYPERTRNS_6
			// setup
			if mo.tics ~= -1
				player.speed = 0
				mo.state = S_PLAY_SUPER_TRANS2
				mo.tics = -1
				mo.frame = $ & FF_TRANSMASK
				chrispy.flags = $ & ~CF_DABBING
				chrispy.supertrans = 2*TICRATE
				chrispy.emeralds = {}
				chrispy.emeraldangle = player.cmd.angleturn*FRACUNIT - ANGLE_90
				mo.height = P_GetPlayerHeight(player)
			end
			
			local frame = mo.frame & FF_FRAMEMASK
			
			if chrispy.supertrans > 0
				// animation
				if not chrispy.transtics
					if frame < C
						chrispy.transtics = 6
					else
						chrispy.transtics = 3
					end
				end
				
				if chrispy.transtics
					chrispy.transtics = $ - 1
					if not chrispy.transtics
						if frame < D
							mo.frame = $ + 1
						else
							mo.frame = $ - 1
						end
					end
				end
				
				// emerald circle!
				// (I detailed the comments here much more thoroughly because another fangame programmer requested to learn how this emerald movement worked!)

				// this code block takes place under a conditional in the game loop that checks that the character is in their transformation animation
				// when the transformation starts, chrispy.supertrans is initialized to 2 seconds (70 frames),
					// chrispy.emeraldangle is set to a base angle value based on the current camera angle,
					// and chrispy.emeralds is set to an empty table which will contain the dummy emerald objects

				local numemeralds = #chrispy.emeralds // (Lua "tables" are all 1-indexed and variable-length arrays; using # returns the length of a table)
				local radius = 3*chrispy.supertrans*mo.scale/2 // we'll use this to determine target coordinates in the for loop

				if chrispy.supertrans % 5 == 0 and numemeralds < 7 // every 5 frames, spawn an emerald, as long as the array contains less than 7!
				and player.bot == CC.BOT_NONE
					local thok = P_SpawnMobjFromMobj(mo, 0, 0, player.height >> 2, MT_THOK) // this spawns a dummy object at the character's location (mo is the player object)
					
					thok.state = S_CEMG1 + numemeralds // (changes the object's state so that it displays as one of the 7 emeralds)
					REmeraldChange(thok) // (I added this function later in development to modify the object further to use Chrispy's new emerald sprites)
					S_StartSound(thok, sfx_s3kb3)
					
					table.insert(chrispy.emeralds, thok) // add the dummy emerald object to the table
				end

				for i, thok in ipairs(chrispy.emeralds) // iterate through each dummy emerald in the table
					// this is the angle our target coordinates for this emerald are to be projected from
					// (the target coordinates are not where the emerald should be, but rather where it's trying to be)
					// the chrispy.supertrans timer changes this angle by -10 degrees each frame (an anticlockwise rotation, since the timer counts down)
					// an angle offset is determined by the emerald number, i, multiplied by 360 degrees divided by 7
					local ang = -FixedAngle(i*360*FRACUNIT/7) - chrispy.supertrans*ANG10 + chrispy.emeraldangle
					
					// these are the target x and y coordinates for this emerald based on the given angle
					// P_ReturnThrust functions are basically cos/sin of an angle, multiplied by a radius
					local goalx = mo.x + P_ReturnThrustX(thok, ang, radius)
					local goaly = mo.y + P_ReturnThrustY(thok, ang, radius)
					
					// now we change the momentum values of the emerald!
					// the emerald starts out at the character's position, so the initial distance to the target coordinates is the largest
					// this makes the initial thrust the strongest--good!
					// I wanted it to feel like the emeralds are ejected from the character, perhaps overshooting a little before settling into position
					// the way the emeralds all start at an indiscernably similar angle offset was actually kind of a good luck draw--
						// if you would prefer not to trial-and-error-adjust your starting angle offset or rate of angle change speed manually,
						// you could keep track of the angular displacement traveled by the most recently spawned emerald and move the
						// emerald-spawning code block into the loop to spawn an emerald each time the angular displacement surpasses 360/7 degrees
					thok.momx = $ + (goalx - thok.x - $)/4
					thok.momy = $ + (goaly - thok.y - $)/4 // ($ represents the variable being modified, i.e. x = $ + 1 is the same as x = x + 1)
					if thok.eflags & MFE_VERTICALFLIP
						thok.z = mo.z + (3*mo.height >> 2) - thok.height // move the z coordinate of the emerald in reverse gravity
					else
						thok.z = mo.z + (mo.height >> 2) // move the z coordinate of the emerald in regular gravity
					end
					REmeraldSpin(thok) // (sprite animation)
					thok.tics = 2 // (SRB2's state duration timer; I force it to 2 here each frame so that if a player disconnects from a netgame and becomes inaccessible or
						// somehow exits transformation, the emeralds despawn quickly on their own)
				end

				chrispy.supertrans = $ - 1 // decrement the timer
			else
				if chrispy.emeralds
					chrispy.emeralds = nil
					S_StartSound(mo, sfx_s3k9f) // RRRRREEEWW!
					/*for player in players.iterate
						if player.mo and FixedHypot(FixedHypot(mo.x - player.mo.x, mo.y - player.mo.y), mo.z - player.mo.z) < FixedMul(RING_DIST, mo.scale)
							P_FlashPal(player, PAL_WHITE, 1)
						end
					end*/
				end
				
				// animation
				if not chrispy.transtics
					if frame > E
						chrispy.transtics = 10
					else
						chrispy.transtics = 3
					end
				end
				
				if chrispy.transtics
				and frame < G
					chrispy.transtics = $ - 1
					if not chrispy.transtics
						mo.frame = $ + 1
					end
				end
				
				// exit condition
				if chrispy.supertrans < -24
					if player.charability == CA_FLY
					and player.cc_gaugeenabled ~= false
						player.powers[pw_tailsfly] = tailsflytics
						player.pflags = $ & ~(PF_JUMPED|PF_NOJUMPDAMAGE|PF_SPINNING|PF_STARTDASH) | PF_THOKKED | PF_CANCARRY
						mo.state = S_PLAY_FLY
						P_SetObjectMomZ(mo, 4*FRACUNIT)
					else
						if CC.UsingSuperAnims(player)
						and not player.isxmomentum
							player.pflags = $ | PF_NOJUMPDAMAGE | PF_THOKKED | PF_JUMPED
						else
							player.pflags = $ & ~(PF_THOKKED|PF_NOJUMPDAMAGE) | PF_JUMPED
						end
						mo.state = S_PLAY_JUMP
					end
				elseif frame > E
					chrispy.supertrans = $ - 1
				end
			end
			
			// if xmom hyper, increment pw_super so that colors are cycled
			if CC.XMomHyper(player)
				player.powers[pw_super] = $ + 1
			end
			
			// skincolor fade
			DoSuperTransColor(player, chrispy.supertrans)
			
			if player.pflags & PF_DIRECTIONCHAR
				player.drawangle = (player.cmd.angleturn << 16) - ANGLE_135
			end
		end
		
		// Emote wheel
		
		if chrispy.flags & CF_EMOTEWHEEL
			emotewheel.scale = $ + FixedMul(FRACUNIT - $, FRACUNIT >> 1)
			
			if chrispy.buttons[BT_JUMP] == 1
				CC.StartEmote(player)
			elseif CC_PressOrHold(emotewheel.right)
				emotewheel.selection = $ % CC.numEmotes + 1
				CC.StartSound(nil, sfx_rmenu1, player)
			elseif CC_PressOrHold(emotewheel.left)
				emotewheel.selection = $ - 1 or CC.numEmotes
				CC.StartSound(nil, sfx_rmenu1, player)
			end
			
			emotewheel.angle = $ + (CC.GetEmoteWheelAngle(emotewheel.selection) - $) / 4
		elseif emotewheel.scale
			emotewheel.scale = FixedMul($ - 1, FRACUNIT >> 1)
		end
		
		if mo.state == S_CHRISPY_ACT
			if mo.sprite2 & ~FF_SPR2SUPER == states[mo.state].frame // oops, we were set into this state without having our sprite2 modified. that's a big nono!
				mo.state = S_PLAY_STND
			else
				local data = CC.emotes[emotewheel.selection]
				local emote = GetWheelEmote(chrispy)
				emote.lastlook = data.id
				activeemote = true
				
				if emotewheel.tics > 0
					emotewheel.tics = $ - 1
					if emotewheel.tics == 0
						emotewheel.tics = data.tick and data.tick(player) or emotewheel.duration
						mo.frame = (($ & FF_FRAMEMASK) + 1) % GetSprite2NumFrames(mo.skin, mo.sprite2)
					end
				end
			end
		end
		
		// Crouch
		
		if chrispy.buttons[crouchbutton] == 1
			if mo.state == S_CHRISPY_CROUCH // leave crouch if already in one
				mo.state = S_PLAY_STND
			elseif (actuallygrounded and player.speed <= 5*mo.scale
				or grounded and mo.momx == 0 and mo.momy == 0)
			and not player.cmd.forwardmove
			and not player.cmd.sidemove
			and player.panim ~= PA_EDGE
			and not (chrispy.flags & (CF_BOUNCING|CF_DEFAULTDANCE))
			and not player.powers[pw_nocontrol]
			and player.playerstate == PST_LIVE
				if mo.state == S_POINTY_PEELOUT
				or player.pflags & PF_STARTDASH
					S_StartSound(mo, sfx_rskid2)
				end
				player.pflags = $ & ~(PF_STARTDASH|PF_SPINNING)
				mo.state = S_CHRISPY_CROUCH
				chrispy.flags = $ & ~CF_DABBING | CF_CROUCHING
				chrispy.buttons[crouchbutton] = $ + 1 // :hackerman:
			end
		end
		
		if mo.state == S_CHRISPY_CROUCH
			player.height = 2*skins[mo.skin].height/3
			
			// character switch
			if chrispy.buttons[dabbutton] == 1
			and player.bot == CC.BOT_NONE
				local swapped = false
				local index = chrispy.lastSwappedPlayerNum
				
				// find a player to swap with
				repeat
					index = ($ + 1) % 32
					
					local player2 = players[index]
					
					if player2 == nil
					or player2.bot == CC.BOT_NONE
					or CC.GetBotLeader(player2) ~= player
					or not player2.chrispy
						continue
					end
					
					local mo2 = player2.mo
					local carry = player2.powers[pw_carry]
					
					if not mo2
					or mo2.state ~= mo.state
					or carry ~= CR_NONE and carry ~= CR_ROLLOUT
						continue
					end
					
					swapped = true
					chrispy.lastSwappedPlayerNum = index
					CC.SwapPlayers(player, player2)
					CC.ExecuteForBots(player, CC.SetStats)
				until index == chrispy.lastSwappedPlayerNum
				
				if swapped
					S_StartSound(nil, sfx_rmenu1, player)
				else
					S_StartSound(nil, sfx_trickf, player)
				end
			end
			
			// emote wheel
			if chrispy.buttons[dancebutton] == 1
				if chrispy.flags & CF_EMOTEWHEEL
					chrispy.flags = $ & ~CF_EMOTEWHEEL
					CC.StartSound(nil, sfx_rmenub, player)
				else
					chrispy.flags = $ | CF_EMOTEWHEEL
					emotewheel.left = INT32_MIN
					emotewheel.right = INT32_MIN
					CC.StartSound(nil, sfx_rmenua, player)
				end
			end
			
			if chrispy.flags & CF_EMOTEWHEEL
				player.pflags = $ | PF_FULLSTASIS
			else
				player.pflags = $ | PF_JUMPSTASIS
			end
			
			AngleStand(player)
		elseif chrispy.flags & CF_CROUCHING
			if chrispy.flags & CF_EMOTEWHEEL
				CC.StartSound(nil, sfx_rmenub, player)
			end
			chrispy.flags = $ & ~(CF_CROUCHING|CF_EMOTEWHEEL)
			player.pflags = $ | PF_JUMPDOWN
			player.height = skins[mo.skin].height
		end
		
		// :DAB: (and hating >:C)
		
		local shakedir = 0
		
		if chrispy.flags & CF_DABBING
			if CC.CanDab(player)
			and chrispy.buttons[dabbutton] ~= 1
			and mo.state ~= S_CHRISPY_DANCE
				if mo.state ~= S_CHRISPY_DAB
					mo.state = S_CHRISPY_DAB
				end
			else
				Undab(player)
			end
		elseif chrispy.buttons[dabbutton] == 1
			if mo.state == S_CHRISPY_HATER
				Undab(player)
			elseif CC.CanDab(player)
				chrispy.supertrans = DABTRANS_TIME
				if player.cc_haterenabled
					if grounded
						mo.state = S_CHRISPY_HATER
					end
				else
					mo.state = S_CHRISPY_DAB
					chrispy.flags = $ | CF_DABBING
					player.pflags = $ & ~PF_CANCARRY
					if not grounded
						player.pflags = $ & ~PF_APPLYAUTOBRAKE
					end
				end
			end
		end
		
		if mo.state == S_CHRISPY_HATER
			if not CC.CanDab(player)
				Undab(player)
			else
				local emote = GetWheelEmote(chrispy)
				emote.lastlook = CC.emotes.hater
				activeemote = true
				AngleStand(player)
				shakedir = -1
				table.insert(haters, player)
				
				// drag
				if player.speed
					CC_Autobrake(player, 5*FRACUNIT/2)
				end
			end
		end
		
		if mo.state == S_CHRISPY_DAB
			// dab transformation (kind of just stole this from my 2.1 version tbh)
			if All7Emeralds(emeralds)
			and player.rings >= 50
			and player.charflags & SF_SUPER
			and (not player.powers[pw_super] or chrispy.supertrans <= 0)
				chrispy.supertrans = $ - 1
				
				if chrispy.supertrans < -19
				elseif chrispy.supertrans < 0
					mo.momz = 0
					DoSuperTransColor(player, chrispy.supertrans)
				elseif chrispy.supertrans == 0 // transform!
					chrispy.emeralds = nil
					P_DoSuperTransformation(player)
					mo.state = S_CHRISPY_DAB
					DoSuperTransColor(player, chrispy.supertrans)
					for i = 1, 32
						local spark = P_SpawnMobjFromMobj(mo, 0, 0, player.height >> 1, MT_CHRISPY_SPARKLE)
						local aiming = P_RandomRange(0, 90)*ANG1
						local angle = P_RandomKey(360)*ANG1
						P_InstaThrust(spark, angle, P_ReturnThrustX(spark, aiming, 8*mo.scale))
						P_SetObjectMomZ(spark, P_ReturnThrustY(spark, aiming, 12*mo.scale))
						spark.flags = $ & ~MF_NOGRAVITY
						spark.frame = P_RandomKey(states[spark.state].var1 + 1) | ($ & ~FF_FRAMEMASK)
						spark.tics = 3*$
						spark.scale = 3*$/2
					end
				elseif chrispy.supertrans <= DABTRANS_THRESHOLD // sequence
					if chrispy.supertrans == DABTRANS_THRESHOLD
						local overlay = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_OVERLAY)
						overlay.skin = mo.skin
						overlay.state = S_CHRISPY_DABTRANS
						overlay.target = mo
						
						chrispy.transoverlay = overlay
						chrispy.emeralds = {}
						chrispy.emeraldangle = player.cmd.angleturn*FRACUNIT - ANGLE_90
					elseif chrispy.supertrans == 12
						local thok = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
						thok.flags2 = $ | MF2_DONTDRAW
						thok.tics = TICRATE
						S_StartSound(thok, sfx_s3k5a)
						for i, thok in ipairs(chrispy.emeralds)
							P_InstaThrust(thok, R_PointToAngle2(thok.x, thok.y, mo.x, mo.y), FixedHypot(thok.x - mo.x, thok.y - mo.y)/12)
						end
					end
					
					local dabtframes = GetSprite2NumFrames(mo.skin, states[S_CHRISPY_DABTRANS].frame)
					
					if chrispy.supertrans < DABTRANS_COLORINTERVAL*16
						DoSuperTransColor(player, chrispy.supertrans / DABTRANS_COLORINTERVAL)
					end
					
					chrispy.transoverlay.frame = $ & ~FF_FRAMEMASK | (dabtframes*(DABTRANS_THRESHOLD - chrispy.supertrans)/DABTRANS_THRESHOLD)
					chrispy.transoverlay.color = player.powers[pw_dye] or player.skincolor
					
					mo.momx = 0
					mo.momy = 0
					mo.momz = 2*flip*mo.scale
					
					if chrispy.supertrans % 12 == 0 and #chrispy.emeralds < 7 // spawn emeralds
						local thok = P_SpawnMobjFromMobj(mo, 0, 0, player.height >> 2, MT_THOK)
						
						thok.state = S_CEMG1 + #chrispy.emeralds
						S_StartSound(thok, sfx_s3kb3)
						
						table.insert(chrispy.emeralds, thok)
					end
					
					for i, thok in ipairs(chrispy.emeralds) // emerald circle!
						thok.tics = 2
						if thok.eflags & MFE_VERTICALFLIP
							thok.z = mo.z + (3*mo.height >> 2) - thok.height
						else
							thok.z = mo.z + (mo.height >> 2)
						end
					end
					
					if chrispy.supertrans > 12
						for i, thok in ipairs(chrispy.emeralds)
							local ang = i*ANG1/7*360 + chrispy.supertrans*ANG2 + chrispy.emeraldangle
							P_TeleportMove(thok,
								mo.x + 64*FixedMul(cos(ang), thok.scale),
								mo.y + 64*FixedMul(sin(ang), thok.scale),
								thok.z + P_RandomRange(-8, 8)*thok.scale)
							REmeraldSpin(thok)
							if not (thok.flags2 & MF2_DONTRESPAWN)
								local radius = FixedDiv(mo.radius, thok.scale)
								local overlay
								
								REmeraldChange(thok)
								
								for i = 1, 8
									P_SpawnMobjFromMobj(thok,
										FixedMul(radius << 1, P_RandomFixed()) - radius,
										FixedMul(radius << 1, P_RandomFixed()) - radius,
										FixedMul(radius << 1, P_RandomFixed()) - 10*FRACUNIT,
										MT_CHRISPY_SPARKLE).momz = mo.momz
								end
								
								overlay = P_SpawnMobjFromMobj(thok, 0, 0, 0, MT_OVERLAY)
								overlay.target = thok.cc_overlay
								overlay.state = S_CHRISPY_REMERALD_BACK
								overlay.frame = ($ + 1) & ~FF_TRANSMASK
								overlay.tics = 2
								thok.flags2 = $ | MF2_DONTRESPAWN
							end
						end
					end
				end
			end
			
			// actual dab stuff
			mo.frame = $ & ~FF_FRAMEMASK
			AngleStand(player)
			shakedir = 1
			if not grounded
				mo.state = $
			end
			table.insert(dabbers, player)
		end
		
		// screenshake calculations
		// applied at end of ThinkFrame function
		
		if shakedir
		and cv_bass.value
		and displayplayer
			if displayplayer.realmo
				local mo = secondarydisplayplayer and secondarydisplayplayer.realmo or mo
				local client = displayplayer.realmo
				local dist = FixedHypot(FixedHypot(mo.x - client.x, mo.y - client.y), mo.z - client.z)
				local maxdist = (player.powers[pw_super] and 2 or 1) * FixedMul(DAB_DISTANCE, mo.scale)
					
				if dist < maxdist
					dabshake = $ + shakedir*FixedMul((player.powers[pw_super] and 3 or 1) * FixedMul(DAB_POWER, mo.scale), FixedDiv(maxdist - dist, maxdist))
				end
			end
		end
		
		// Tpose
		
		if mo.state == S_CHRISPY_TPOSE
			player.drawangle = (player.cmd.angleturn << 16) + ANGLE_135
		end
		
		// High Five
		
		if mo.state == S_CHRISPY_HIGHFIVE
			HandleHighFive(player)
			
			if chrispy.highFiveState == 0
				local emote = GetWheelEmote(chrispy)
				emote.lastlook = CC.emotes.highfive
				activeemote = true
				AngleStand(player)
			end
		elseif CC.ActiveHighFiveState(player)
			chrispy.highFiveState = 0
			CC.StartSound(nil, sfx_trickf, player)
		end
		
		// Teeter foot switch
		
		if mo.state == S_PLAY_EDGE
			if chrispy.prevstate ~= S_PLAY_EDGE and P_RandomChance(FRACUNIT/2)
			or chrispy.buttons[crouchbutton] == 1
				mo.mirrored = not $
				for i = 1, GetSprite2NumFrames(mo.skin, states[S_PLAY_EDGE].frame)/2
					mo.state = $
				end
			end
		elseif chrispy.prevstate == S_PLAY_EDGE
			mo.mirrored = false
		end
		
		// Drop Dash
		
		if chrispy.buttons[BT_USE] == 1
		and CC.CanUseDropDash(player)
			chrispy.dropdash = -abs(DROPDASH_DELAY)
			chrispy.flags = $ | CF_DROPDASHCHARGE
		end
		
		if actuallygrounded and not chrispy.grounded
			if chrispy.flags & CF_DROPDASHCHARGE
			and chrispy.dropdash > 0
			and chrispy.buttons[BT_USE]
			and player.playerstate == PST_LIVE
			and not P_PlayerInPain(player)
				local charge = min(DROPDASH_MINCHARGE + chrispy.dropdash*DROPDASH_CHARGESPEED, DROPDASH_MAXCHARGE)
				player.drawangle = mo.angle
				CC.SpawnDust(player, FixedMul(FixedDiv(charge, player.maxdash), mo.scale), mo.scale >> 2, 0, 6)
				charge = FixedMul($, mo.scale)
				if mo.eflags & MFE_UNDERWATER
					charge = $ >> 1
				end
				if player.powers[pw_super]
					charge = 4*$/3
					if player == displayplayer
					or player == displayplayer
						P_StartQuake(8*mo.scale, 8, nil, 256*mo.scale)
					end
					S_StartSound(mo, sfx_pstop)
				end
				P_InstaThrust(mo, mo.angle, charge)
				S_StartSound(mo, sfx_zoom)
				player.pflags = $ | PF_SPINNING
				mo.state = S_PLAY_ROLL
			end
			chrispy.flags = $ & ~CF_DROPDASHMASK
		end
		
		if chrispy.flags & CF_DROPDASHCHARGE
			if not chrispy.buttons[BT_USE]
			or not CC.CanUseDropDash(player)
				if (player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE)) == PF_JUMPED
					mo.state = S_PLAY_JUMP
					if not (chrispy.flags & CF_USEDDROPDASH)
						player.pflags = $ & ~PF_THOKKED
					end
				end
				chrispy.flags = $ & ~CF_DROPDASHCHARGE | CF_USEDDROPDASH
				chrispy.ddtrans = DROPDASH_DELAY - abs($)
			else
				local charge = DROPDASH_MINCHARGE + chrispy.dropdash*DROPDASH_CHARGESPEED
				chrispy.dropdash = $ + 1
				if chrispy.dropdash > 0
					player.drawangle = mo.angle
					if charge >= DROPDASH_MAXCHARGE
						if charge == DROPDASH_MAXCHARGE
							S_StartSound(mo, sfx_cdfm40)
						end
						mo.state = S_PLAY_SPINDASH
					elseif chrispy.dropdash == 1
						S_StartSound(mo, sfx_ddash)
						player.pflags = $ & ~PF_SPINNING | PF_THOKKED
						mo.state = S_CHRISPY_DROPDASH
					elseif mo.state ~= S_CHRISPY_DROPDASH
						chrispy.flags = $ & ~CF_DROPDASHCHARGE | CF_USEDDROPDASH
						chrispy.ddtrans = DROPDASH_DELAY - abs($)
					end
				else
					chrispy.ddtrans = chrispy.dropdash - 1
				end
			end
		elseif chrispy.ddtrans
			if mo.sprite2 ~= SPR2_ROLL
			or grounded
				chrispy.ddtrans = 0
			elseif chrispy.ddtrans < 0
				chrispy.ddtrans = $ + 1
			else
				chrispy.ddtrans = $ - 1
			end
		end
		
		local trans = chrispy.ddtrans
		if trans < 0 // transition to dropdash angle
			trans = abs($)
			player.drawangle = mo.angle + trans * (($ - mo.angle) / DROPDASH_DELAY)
		end
		
		// toggle SF_NOSHIELDABILITY
		
		CheckShieldAbility(player, grounded)
		
		// High-speed roll
		
		local speed = FixedDiv(FixedHypot(mo.momx - player.cmomx, mo.momy - player.cmomy), mo.scale)
		local normalspeed = skins[mo.skin].normalspeed
		
		if speed < normalspeed
			if chrispy.flags & CF_FASTROLL
				if player.pflags & (PF_STARTDASH|PF_SPINNING) == PF_SPINNING
				and mo.state == S_PLAY_SPINDASH
					mo.state = S_PLAY_ROLL
				end
				chrispy.flags = $ & ~CF_FASTROLL
			end
		elseif speed > normalspeed + CC_SPEED_INCREASE_LEEWAY // value from POINTY-SONIC.lua
			chrispy.flags = $ | CF_FASTROLL
		end
		
		if chrispy.flags & CF_FASTROLL
		and player.pflags & (PF_SPINNING|PF_JUMPED) == PF_SPINNING
			mo.state = S_PLAY_SPINDASH
		end
		
		// midfall frame
		
		if mo.state == S_PLAY_SPRING
		and not (mo.sprite2 & FF_SPR2SUPER)
			if flip*mo.momz < 8471*mo.scale/5000 // ( ͡º ͜ʖ ͡º) see if you can crack why this is funny
				mo.state = S_CHRISPY_MIDFALL
			end
		elseif mo.state == S_CHRISPY_MIDFALL
			if -flip*mo.momz > 8471*mo.scale/5000
				mo.state = S_PLAY_FALL
			end
		end
		
		// Fan frame
		
		if not grounded
		and (player.powers[pw_carry] == CR_NONE or player.powers[pw_carry] == CR_DUSTDEVIL)
		and not player.powers[pw_tailsfly]
		and not (player.pflags & (PF_SPINNING|PF_SLIDING|PF_GLIDING|PF_BOUNCING))
		and (player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE)) ~= PF_JUMPED
		and not (chrispy.flags & CF_DABBING)
		and mo.state ~= S_CHRISPY_TRICK
		and mo.state ~= S_POINTY_FLIP
		and player.playerstate == PST_LIVE
		and not P_PlayerInPain(player)
			if player.powers[pw_carry] == CR_DUSTDEVIL
				chrispy.fantics = -1
			end
			if chrispy.fantics
				mo.state = S_CHRISPY_FAN
				mo.tics = chrispy.fantics
				chrispy.fantics = 0
			elseif P_PlayerTouchingSectorSpecial(player, 4, 5) // Fan sectors
				if mo.state ~= S_CHRISPY_FAN
					mo.state = S_CHRISPY_FAN
				end
			elseif not (CC.UsingSuperAnims(player)) // current sectors - this is more involved...
				local currents = {}
				table.insert(currents, P_PlayerTouchingSectorSpecial(player, 3, 2))
				table.insert(currents, P_PlayerTouchingSectorSpecial(player, 3, 3))
				
				for _, sector in ipairs(currents)
					if not sector.valid
						continue
					end
					
					if fansectors[#sector] ~= nil
						if mo.state ~= S_CHRISPY_FAN
						and fansectors[#sector] == true
							mo.state = S_CHRISPY_FAN
						end
						break
					end
					
					local val = false
					
					for j = 541, 546 // wind & current linedef specials
						if P_FindSpecialLineFromTag(j, sector.tag, -1) ~= -1 // I'm gonna assume that if a line is found then the effect is deliberate
							val = true
							break
						end
					end
					
					fansectors[#sector] = val
					if val
					and mo.state ~= S_CHRISPY_FAN
						mo.state = S_CHRISPY_FAN
					end
				end
			end
		end
		
		if mo.state == S_CHRISPY_FAN
			if (CC.UsingSuperAnims(player))
				mo.sprite2 = SPR2_RUN|FF_SPR2SUPER
			end
			if chrispy.fanangle == nil
				chrispy.fanangle = player.drawangle
			else
				chrispy.fanangle = $ + states[mo.state].var1
			end
			player.drawangle = chrispy.fanangle
		elseif chrispy.fanangle ~= nil
			chrispy.fanangle = nil
		end
		
		// menu emote
		
		if player.cc_menu
		and player.cc_menu.active
			local emote = GetWheelEmote(chrispy)
			emote.lastlook = CC.emotes.thinking
			chrispy.flags = $ & ~CF_EMOTEWHEEL
			activeemote = true
			if mo.state == S_PLAY_LOOK
				mo.state = S_PLAY_WAIT
				CC_KillNeck(mo.neck)
			end
		end
		
		// Spindash visuals
		
		if player.pflags & (PF_STARTDASH|PF_JUMPED) == PF_STARTDASH
			if mo.sprite2 ~= SPR2_SPIN
				if not (leveltime % 4)
					local orb = P_SpawnMobjFromMobj(mo, 0, 0, chrispy.revheight*FRACUNIT, player.spinitem)
					orb.target = mo
					orb.state = S_CHRISPY_REVORB
					orb.color = mo.color
					orb.frame = $ + leveltime/8%3
				end
				
				if not (leveltime & 1)
					local hangle = FixedAngle(P_RandomKey(360)*FRACUNIT)
					local vangle = FixedAngle(P_RandomRange(-180, 180)*FRACUNIT)
					local dist = 64*FRACUNIT
					local thrust = FixedMul(cos(vangle), dist)
					local orb = P_SpawnMobjFromMobj(mo,
						FixedMul(cos(hangle), thrust),
						FixedMul(sin(hangle), thrust),
						FixedMul(sin(vangle), dist) + (flip*mo.height >> 1),
						player.spinitem)
					orb.target = mo
					orb.scale = mo.scale/5
					orb.color = mo.color
					orb.tics = 10
					table.insert(chrispy.spinparticles, orb)
				end
			elseif grounded and player.dashspeed > player.mindash
				CC.SpawnDust(player, FixedMul(FixedDiv(player.dashspeed, player.maxdash), mo.scale), mo.scale >> 2, 0, 4)
			end
		end
		
		for i = #chrispy.spinparticles, 1, -1
			local orb = chrispy.spinparticles[i]
			if not valid(orb)
				table.remove(chrispy.spinparticles, i)
			else
				local x = orb.x + ((mo.x - orb.x) / 8)
				local y = orb.y + ((mo.y - orb.y) / 8)
				local z = orb.z + ((mo.z + (mo.height >> 1) - orb.z) / 8)
				P_TeleportMove(orb, x, y, z)
			end
		end
		
		// Tricks highscore cash!
		
		local tricks = mo.cc_tricks
		
		if tricks and tricks.count > player.cc_tricks_highscore
			local radius = FixedDiv(mo.radius, mo.scale) << 1
			local height = FixedDiv(mo.height, mo.scale) << 1
			while P_RandomChance(FRACUNIT/3)
				local coin = P_SpawnMobjFromMobj(mo,
					FixedMul(radius << 1, P_RandomFixed()) - radius,
					FixedMul(radius << 1, P_RandomFixed()) - radius,
					FixedMul(height, P_RandomFixed()), MT_CHRISPY_COIN)
				coin.target = mo
				S_StartSound(coin, coin.info.seesound)
			end
			CC_SpawnConfetti(player, 1, 12*FRACUNIT, nil, false, function(part)
				A_SetObjectFlags(part, MF_NOBLOCKMAP, 1)
				part.state = S_CHRISPY_CASH
			end)
			if not S_SoundPlaying(mo, sfx_trckhs)
				S_StartSound(mo, sfx_trckhs)
			end
		end
		
		// Super sparkles & Fleet Feet
		
		if mo.state >= S_PLAY_SUPER_TRANS1
		and mo.state <= S_PLAY_SUPER_TRANS6
		and chrispy.supertrans > 0
			// ;
		elseif player.powers[pw_super]
			local radius = FixedDiv(mo.radius, mo.scale)
			local height = FixedDiv(mo.height, mo.scale)
			P_SpawnMobjFromMobj(mo,
				FixedMul(2*radius, P_RandomFixed()) - radius,
				FixedMul(2*radius, P_RandomFixed()) - radius,
				FixedMul(height + 4*FRACUNIT, P_RandomFixed()),
				MT_CHRISPY_SPARKLE)
		elseif player.powers[pw_sneakers]
			local radius = FixedDiv(mo.radius, mo.scale)
			P_SpawnMobjFromMobj(mo,
				FixedMul(2*radius, P_RandomFixed()) - radius,
				FixedMul(2*radius, P_RandomFixed()) - radius,
				FixedMul(10*FRACUNIT, P_RandomFixed()) - 2*FRACUNIT,
				MT_CHRISPY_SPARKLE)
		end
		
		// Hyper colors
		
		if CC.GodModeHyper(player)
		and mo.color <= SKINCOLOR_SUPERPURPLE5
		and mo.color >= FIRSTSUPERCOLOR
			mo.color = ($ - FIRSTSUPERCOLOR)/5 + SKINCOLOR_HYPERPOINTYSILVER + (skins[mo.skin].supercolor - SKINCOLOR_SUPERLIGHTBULB1)
			if mo.neck
				for i, head in ipairs(mo.neck.heads)
					head.color = mo.color
				end
			end
		end
		
		// xmom faceplant
		
		if player.isxmomentum
		and mo.state == S_PLAY_FACEPLANT
		and mo.sprite2 & ~FF_SPR2SUPER ~= SPR2_PDED
			mo.sprite2 = SPR2_PDED
			mo.frame = (GetSprite2NumFrames(mo.skin, states[S_CHRISPY_DEATH].frame) - 1) | ($ & ~FF_FRAMEMASK)
		end
		
		// Sound effects (footsteps)
		
		local helpersound
		local helpertics
		
		if (cv_steps.value or mo.sprite2 == SPR2_FLY and player.pflags & PF_CANCARRY) // exceptions suuuck
		and CC_Footsteps[mo.skin]
			local sounds = CC_Footsteps[mo.skin][mo.sprite2] or CC_Footsteps[mo.skin][mo.sprite2 & ~FF_SPR2SUPER]
			local flags = sounds and sounds.flags or 0
			local frame = mo.frame & FF_FRAMEMASK
			local waterheight
			if mo.eflags & MFE_VERTICALFLIP
				waterheight = mo.waterbottom
			else
				waterheight = mo.watertop
			end
			
			if sounds
			and (grounded and mo.z ~= waterheight or flags & FSF_AERIAL)
				local prevframe = chrispy.prevstep & (FRACUNIT - 1)
				local prevsprite2 = chrispy.prevstep >> FRACBITS
				local sound = sounds[frame]
				
				if sounds.repeatsound
				and (not (flags & FSF_MOVEMENTONLY) or player.speed > mo.scale)
					helpersound = sounds.repeatsound & (FRACUNIT - 1)
					helpertics = sounds.repeatsound >> FRACBITS
				end
				
				if sound
				and (prevsprite2 ~= mo.sprite2 and not (flags & FSF_NOTRANSITION)
				or prevsprite2 == mo.sprite2 and prevframe ~= frame)
					while sound & (FRACUNIT - 1)
						S_StartSound(mo, sound & (FRACUNIT - 1))
						sound = $ >> FRACBITS
					end
				end
			end
			chrispy.prevstep = (mo.sprite2 << FRACBITS) | frame
			
			// repeatsounds
			local helper = chrispy.soundhelper
			if not valid(helper) // spawn helping object
				helper = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
				helper.state = S_INVISIBLE
				helper.skin = mo.skin
				chrispy.soundhelper = helper
			end
			P_TeleportMove(helper, mo.x, mo.y, mo.z)
			
			if helper.sound ~= helpersound
				S_StopSound(helper)
				helper.sound = helpersound
				helper.tics = helpertics or -1
			end
			
			if helper.sound
			and not S_SoundPlaying(helper, helper.sound)
				S_StartSound(helper, helper.sound)
			end
		end
		
		// trick reset
		/*if grounded and not chrispy.grounded
			chrispy.flags = $ & ~CF_FIRSTBOUNCE
		end*/
		
		// kill headtracking objects if no longer looking
		if looking and mo.state ~= S_PLAY_LOOK
			CC_KillNeck(mo.neck)
		end
		
		// kill emote if nothing is using it
		
		if not activeemote
		and valid(chrispy.emote)
			chrispy.emote.target = nil
		end
		
		// resets
		chrispy.grounded = grounded
		chrispy.prevstate = mo.state
		chrispy.prevmomentum = {mo.momx, mo.momy, mo.momz}
	end
	
	// Screenshake/noise
	
	if dabshake > 0
		P_StartQuake(dabshake, 1)
	end
	
	// confetti reduction
	
	for _, player in ipairs(dabbers)
		local numhaters = 0
		local numdabbers = 0
		
		for _, dabber in ipairs(dabbers)
			local dist = FixedHypot(FixedHypot(player.mo.x - dabber.mo.x, player.mo.y - dabber.mo.y),
				player.mo.z - dabber.mo.z)
			local maxdist = (player.powers[pw_super] and 2 or 1) * FixedMul(DAB_DISTANCE, dabber.mo.scale)
			if dist < maxdist
				numdabbers = $ + 1
			end
		end
		
		for _, hater in ipairs(haters)
			local dist = FixedHypot(FixedHypot(player.mo.x - hater.mo.x, player.mo.y - hater.mo.y),
				player.mo.z - hater.mo.z)
			local maxdist = (player.powers[pw_super] and 2 or 1) * FixedMul(DAB_DISTANCE, hater.mo.scale)
			if dist < maxdist
				numhaters = $ + 1
			end
		end
		
		if numdabbers >= numhaters
			CC_SpawnConfetti(player, 1, 12*FRACUNIT)
			if cv_bass.value
				S_StartSoundAtVolume(player.mo, sfx_dab, 255/(numhaters + 1))
			end
		end
	end
end)

// Allow dropdash on first spin press

CC.AddHook("ShieldSpecial", function(player)
	local chrispy = player.chrispy
	
	if not chrispy
	or player.gotflag or player.gotcrystal
		return
	end
	
	if player.powers[pw_shield] & SH_NOSTACK == SH_WHIRLWIND
	and player.pflags & PF_JUMPED
		P_DoJumpShield(player)
		player.pflags = $ & ~PF_NOJUMPDAMAGE
		player.mo.state = S_PLAY_ROLL
		return true
	end
end)

// Set player height when using dropdash, bounce, etc.

local function DoHeight(player)
	local chrispy = player.chrispy
	if not chrispy
		return
	end
	
	local mo = player.mo
	
	if mo.state == S_POINTY_BOUNCE_LANDING
		return P_GetPlayerSpinHeight(player) >> 1
	end
	
	if chrispy.flags & (CF_DROPDASHCHARGE|CF_BOUNCING)
		return P_GetPlayerSpinHeight(player)
	end
end

CC.AddHook("PlayerHeight", DoHeight)
CC.AddHook("PlayerCanEnterSpinGaps", DoHeight)

// object thinkers

CC.AddHook("MobjSpawn", function(spark) // we got our own super sparks, fuck yours
	if not spark.valid
		return
	end
	
	local bestmo
	local bestdist = INT32_MAX
	for player in players.iterate
		local mo = player.mo
		if mo
			local dist = FixedHypot(mo.x - spark.x, mo.y - spark.y)
			if dist < bestdist
				bestdist = dist
				bestmo = mo
			end
		end
	end
	if IsChrispySkin(bestmo.skin)
	and (bestmo.state ~= S_POINTY_BOUNCE_LANDING)
		spark.state = S_INVISIBLE
		spark.fuse = 1
	end
end, MT_SUPERSPARK)

CC.AddHook("MobjThinker", function(coin)
	if P_IsObjectOnGround(coin)
	and coin.reactiontime
		coin.reactiontime = $ - 1
		S_StartSound(coin, coin.info.painsound)
	end
end, MT_CHRISPY_COIN)

CC.AddHook("MobjThinker", function(mo)
	if mo.state == S_CHRISPY_BUBBLEDUST
	and not (mo.eflags & MFE_UNDERWATER)
		P_RemoveMobj(mo)
		return
	end
	
	if leveltime % 3 == 0
		local trans = max((mo.frame & FF_TRANSMASK) + FF_TRANS10, 0)
		if trans > FF_TRANS90
			P_KillMobj(mo)
			return
		end
		mo.frame = (mo.frame & ~FF_TRANSMASK) | trans
	end
	P_Thrust(mo, mo.angle, (FixedHypot(mo.momx, mo.momy) >> 4))
end, MT_CHRISPY_DUST)

local function DoSpinTrail(mo)
	if not (mo.flags2 & MF2_DONTRESPAWN)
		mo.interpmode = 1 // scaling doesn't interpolate as of writing but this still looks best
		mo.flags2 = $ | MF2_DONTRESPAWN
		mo.destscale = mo.scale >> 3
		mo.scalespeed = (mo.scale - mo.destscale)/mo.tics
	end
	mo.frame = ($ & ~FF_TRANSMASK) | min($ & FF_TRANSMASK + FF_TRANS10, FF_TRANS90)
end

CC.AddHook("MobjThinker", function(mo)
	if mo.valid
		DoSpinTrail(mo)
	end
end, MT_POINTY_SPINTRAIL)

CC.AddHook("MobjThinker", function(mo)
	if mo.valid
		DoSpinTrail(mo)
		if mo.tics == 3
			mo.color = SKINCOLOR_WHITE
		end
	end
end, MT_FLUFFY_SPINTRAIL)

// Fan frame bullshit

local function DoFanAndGasJet(mo, spring) // port of P_DoFanAndGasJet from p_map.c
	if not (mo.valid and spring.valid)
	or (spring.type ~= MT_FAN and spring.type ~= MT_STEAM)
		return
	end
	
	local player = mo.player
	local chrispy = player.chrispy
	local zdist
	local speed = spring.info.mass
	local flip = P_MobjFlip(spring)
	
	if not (player and chrispy)
	or P_PlayerInPain(player)
		return
	end
	
	if spring.eflags & MFE_VERTICALFLIP
		if mo.z > spring.z + spring.height
			return
		end
		zdist = (spring.z + spring.height) - (mo.z + mo.height)
	else
		if mo.z + mo.height < spring.z
			return
		end
		zdist = mo.z - spring.z
	end
	
	if spring.type == MT_FAN
		if zdist > spring.health << FRACBITS
		or flip*mo.momz >= FixedMul(speed, spring.scale)
		or player.climbing or player.pflags & PF_GLIDING
			return
		end
		
		chrispy.fantics = max($, 1)
	else
		if zdist > 16*spring.scale
		or spring.state ~= S_STEAM1
			return
		end
		
		chrispy.fantics = max($, 720/abs(states[S_CHRISPY_FAN].var1/ANG1))
	end
end

CC.AddHook("MobjMoveCollide", DoFanAndGasJet, MT_PLAYER)
CC.AddHook("MobjCollide", DoFanAndGasJet, MT_PLAYER)

// Get death element

CC.AddHook("MobjDeath", function(mo, inflictor, source, damagetype)
	if not (mo.valid and mo.chrispy) return end
	if valid(inflictor) and inflictor.flags & MF_FIRE
		mo.chrispy.deathtype = DMG_FIRE
	else
		mo.chrispy.deathtype = damagetype
	end
end, MT_PLAYER)

// Signpost finger wag!

local signs = {}

CC.AddHook("MobjSpawn", function(mo)
	table.insert(signs, mo)
end, MT_SIGN)

CC.AddHook("ThinkFrame", do
	for i = #signs, 1, -1
		local sign = signs[i]
		
		if not sign.valid
		or not (sign.tracer and sign.tracer.tracer)
			table.remove(signs, i)
			continue
		end
		
		local face = sign.tracer.tracer
		local mo = sign.target
		local player = valid(mo) and mo.player
		local skin = face.skin
		
		if IsChrispySkin(skin)
			local function SetupFinger(finger)
				finger.skin = skin
				finger.frame = FF_PAPERSPRITE
				finger.angle = face.angle
				finger.threshold = 1
				face.chrispy_finger = finger
				
				if player
					if player.rings == 1 // haha, ha, one!
					and skin == "pointy"
						finger.frame = $ | B
					elseif skin == "fluffy"
						finger.frame = $ | C
					end
					
					// :pointysass:
					if player.realtime / TICRATE == 4*60 + 20
						for i = 1, 2
							face.state = S_PLAY_SIGN
						end
					elseif player.rings == 69
						face.state = S_PLAY_SIGN
					end
				end
				face.tics = -1
			end
			
			local finger = face.chrispy_finger
			if valid(finger)
				if skin ~= finger.skin
					SetupFinger(finger)
				end
			else
				finger = P_SpawnMobjFromMobj(face, 0, 0, 0, MT_OVERLAY)
				finger.target = face
				finger.state = S_INVISIBLE
				finger.sprite = SPR_ARM_
				SetupFinger(finger)
			end
			
			finger.color = face.color
			if face.target.color == SKINCOLOR_YELLOW
				face.target.color = SKINCOLOR_SUPERGOLD5
			end
			
			if CC_GetCVar(CC_Renderer) == 1
				finger.rollangle = $ + finger.threshold * (ANG2)
				if finger.rollangle > ANG20
				or finger.rollangle < -ANG10
					finger.threshold = -$
				end
			else
				finger.rollangle = 0
			end
		elseif valid(face.chrispy_finger)
			P_RemoveMobj(face.chrispy_finger)
			face.chrispy_finger = nil
		end
	end
end)

CC.AddHook("NetVars", function(net)
	signs = net(signs)
end)

// Emote wheel drawer

local function R_GetScreenCoords(v, p, c, mx, my, mz) // thanks to Lat' for letting me nab this from SRB2P 1.2
	local camx, camy, camz, camangle, camaiming
	if p.awayviewtics then
		camx = p.awayviewmobj.x
		camy = p.awayviewmobj.y
		camz = p.awayviewmobj.z
		camangle = p.awayviewmobj.angle
		camaiming = p.awayviewaiming
	elseif c.chase then
		camx = c.x
		camy = c.y
		camz = c.z
		camangle = c.angle
		camaiming = c.aiming
	else
		camx = p.mo.x
		camy = p.mo.y
		camz = p.viewz-20*FRACUNIT
		camangle = p.mo.angle
		camaiming = p.aiming
	end

	-- Lat: I'm actually very lazy so mx can also be a mobj!
	if type(mx) == "userdata" and mx.valid
		my = mx.y
		mz = mx.z
		mx = mx.x	-- life is easier
	end

	local x = camangle-R_PointToAngle2(camx, camy, mx, my)

	local distfact = cos(x)
	if not distfact then
		distfact = 1
	end -- MonsterIestyn, your bloody table fixing...

	if x > ANGLE_90 or x < ANGLE_270 then
		return -9, -9, 0
	else
		x = FixedMul(tan(x, true), 160<<FRACBITS)+160<<FRACBITS
	end

	local y = camz-mz
	--print(y/FRACUNIT)
	y = FixedDiv(y, FixedMul(distfact, R_PointToDist2(camx, camy, mx, my)))
	y = (y*160)+(100<<FRACBITS)
	y = y+tan(camaiming, true)*160

	local scale = FixedDiv(160*FRACUNIT, FixedMul(distfact, R_PointToDist2(camx, camy, mx, my)))
	--print(scale)

	return x, y, scale
end

local function EmoteWheelHUD(v, player, cam)
	local mo = player.mo
	local emotewheel = player.chrispy.emotewheel
	
	if emotewheel.scale > FRACUNIT >> 6
		local flip = 1
		local bubble = v.cachePatch("ROPTBUBL")
		local angdiff = FixedAngle(360*FRACUNIT/CC.numEmotes)
		local x, y, scale
		local cutoff = function(y) return false end
		
		if cam.chase and not player.awayviewtics
			x, y, scale = R_GetScreenCoords(v, player, cam, mo)
			if mo.eflags & MFE_VERTICALFLIP
			and player.pflags & PF_FLIPCAM
				y = 200*FRACUNIT - $
			else
				flip = P_MobjFlip(mo)
			end
		else
			x, y, scale = 160*FRACUNIT, (100 + bubble.height >> 1)*FRACUNIT, FRACUNIT
		end
		
		scale = FixedMul($, emotewheel.scale)
		
		if splitscreen
			if player == secondarydisplayplayer
				cutoff = function(y) return y < (bubble.height*scale >> 1) end
			else
				cutoff = function(y) return y > 200*FRACUNIT + (bubble.height*scale >> 1) end
			end
		end
		
		for i = 1, CC.numEmotes
			local emote = CC_NeckSkins["pointy"].trackables[CC.emotes[i].id]
			local patch
			local angle = (i - 1)*angdiff + ANGLE_90 - emotewheel.angle
			local x = x - P_ReturnThrustX(nil, angle, 64*scale)
			local y = y - flip*P_ReturnThrustY(nil, angle, 64*scale)
			
			if cutoff(y) continue end
			
			if emote.emotesprite == SPR_PLAY
				patch = v.getSprite2Patch(emote.emoteskin or mo.skin, emote.emotesprite2 or SPR2_STND, emote.emoteframe)
			else
				patch = v.getSpritePatch(emote.emotesprite, emote.emoteframe)
			end
			
			v.drawScaled(x, y, scale, bubble, V_PERPLAYER,
				v.getColormap(TC_DEFAULT, i == emotewheel.selection and ColorOpposite(player.skincolor) or player.skincolor))
			v.drawScaled(x, y, scale, patch, V_PERPLAYER,
				v.getColormap(TC_DEFAULT, emote.emotecolor or SKINCOLOR_NONE))
		end
	end
end

local wheelPlayers = {}

local function SortWheelPlayers(a, b)
	a = a.mo
	b = b.mo
	
	return R_PointToDist(a.x, a.y) < R_PointToDist(b.x, b.y)
end

hud.add(function(v, p, cam)
	for player in players.iterate
		local onScreen = player == p or player.bot ~= CC.BOT_NONE and CC.GetBotLeader(player) == p
		
		if not onScreen
		or not player.mo
		or not player.chrispy
		or not player.chrispy.emotewheel
			continue
		end
		
		table.insert(wheelPlayers, player)
	end
	
	table.sort(wheelPlayers, SortWheelPlayers)
	
	for i = #wheelPlayers, 1, -1
		EmoteWheelHUD(v, table.remove(wheelPlayers, i), cam)
	end
end)

// cache NiGHTS frames to reduce lag during gameplay
// not necessary in 2.2.9+

if not (MODID == 18 and SUBVERSION < 9)
	return
end

local cachepatch

local function CacheNights(v)
	if not valid(cachepatch)
	and (maptol & TOL_NIGHTS or titlemapinaction)
		local angnames = {[3] = "L", [7] = "R"}
		for _, ang in ipairs({3, 7}) // L/R
			local frame = 0
			while v.getSpritePatch(SPR_NIGH, frame, ang, ANGLE_45)
				print("Rotated patches for NIGH"..R_Frame2Char(frame)..angnames[ang].." cached.")
				frame = $ + 1
			end
		end
		cachepatch = v.cachePatch("MISSING")
	end
end

hud.add(CacheNights, "title")
hud.add(CacheNights, "titlecard")